import base64
import json
import mimeparse
from pyld import jsonld
import re
import requests
import requests_cache
import six
import sys
import uuid

from ..actions.graph import add_node, patch_node, patch_node_reference
from ..actions.input import store_original_resource
from ..actions.validation_report import set_validation_subject
from ..actions.tasks import add_task, delete_outdated_node_tasks, report_message
from ..actions.validation_report import set_openbadges_version
from ..exceptions import TaskPrerequisitesError
from ..openbadges_context import OPENBADGES_CONTEXT_V2_URI
from ..reducers.graph import get_next_blank_node_id
from ..state import get_node_by_id, node_match_exists
from ..utils import list_of, jsonld_use_cache,make_string_from_bytes, MESSAGE_LEVEL_WARNING

from .task_types import (DETECT_AND_VALIDATE_NODE_CLASS, FETCH_HTTP_NODE, INTAKE_JSON, JSONLD_COMPACT_DATA,
                         PROCESS_BAKED_RESOURCE, UPGRADE_0_5_NODE, UPGRADE_1_0_NODE, UPGRADE_1_1_NODE,
                         VALIDATE_EXPECTED_NODE_CLASS, VALIDATE_EXTENSION_NODE)
from .utils import abbreviate_node_id as abv_node, filter_tasks, is_iri, is_url, task_result, URN_REGEX
from .validation import OBClasses


def fetch_http_node(state, task_meta, **options):
    url = task_meta['url']
    depth = task_meta['depth']

    if options.get('cache_backend'):
        session = requests_cache.CachedSession(
            backend=options['cache_backend'], expire_after=options.get('cache_expire_after', 300))
    else:
        session = requests.Session()

    result = session.get(
        url, headers={'Accept': 'application/ld+json, application/json, image/png, image/svg+xml'}
    )

    try:
        json_body = result.json()
        response_text_with_proper_encoding = json.dumps(json_body)
    except ValueError:
        content_type = result.headers.get('Content-Type', 'UNKNOWN')

        if mimeparse.quality(content_type, 'image/svg+xml') > 0.9:
            parsed_type = 'image/svg+xml'
        elif mimeparse.quality(content_type, 'image/png') > 0.9:
            parsed_type = 'image/png'
        else:
            return task_result(
                success=False,
                message="Unknown Content-Type (Not image/png or image/svg+xml). Response could not be interpreted from url {}".format(
                    url)
            )

        b64content = b''.join([b'data:', parsed_type.encode(), b';base64,', base64.b64encode(result.content)])
        actions = [store_original_resource(node_id=url, data=b64content)]
        if task_meta.get('is_potential_baked_input', False):
            actions += [add_task(PROCESS_BAKED_RESOURCE, node_id=url, depth=depth)]

        return task_result(
            True, 'Successfully fetched image from {}'.format(url), actions)

    actions = [
        store_original_resource(node_id=url, data=response_text_with_proper_encoding),
        add_task(INTAKE_JSON,
                 data=response_text_with_proper_encoding,
                 node_id=url,
                 expected_class=task_meta.get('expected_class'),
                 source_node_path=task_meta.get('source_node_path'),
                 depth=depth
                 )
    ]
    return task_result(message="Successfully fetched JSON data from {}".format(url), actions=actions)


def _detect_openbadges_version(data):
    context = data.get('@context')
    if isinstance(context, six.string_types):
        if 'v1' in context:
            return '1.1'
        elif 'v2' in context:
            return '2.0'
    elif context is None:
        if isinstance(data.get('recipient'), six.string_types):
            return '0.5'
        return '1.0'

    return '2.0'  # Default to latest version if we cannot interpret version


def intake_json(state, task_meta, **options):
    input_data = task_meta['data']
    node_id = task_meta.get('node_id')
    expected_class = task_meta.get('expected_class')
    depth = task_meta.get('depth')
    openbadges_version = None
    actions = []

    try:
        data = json.loads(make_string_from_bytes(input_data))
    except TypeError as e:
        return task_result(False, "Could not load JSON from data: " + str(e))

    openbadges_version = _detect_openbadges_version(data)
    actions.append(set_openbadges_version(openbadges_version))

    if openbadges_version in ['1.1', '2.0']:
        compact_action = add_task(
            JSONLD_COMPACT_DATA,
            node_id=node_id,
            openbadges_version=openbadges_version,
            expected_class=expected_class,
            data=input_data,
            source_node_path=task_meta.get('source_node_path'),
            depth=depth
        )
        actions.append(compact_action)

    if openbadges_version == '2.0':
        pass
    elif openbadges_version == '1.1':
        actions.append(add_task(
            UPGRADE_1_1_NODE, node_id=node_id, expected_class=expected_class,
            prerequisites=[compact_action['task_key']]
        ))
    elif openbadges_version == '1.0':
        actions.append(add_task(
            UPGRADE_1_0_NODE, node_id=node_id, expected_class=expected_class, data=input_data,
            source_node_path=task_meta.get('source_node_path')
        ))
    elif openbadges_version == '0.5':
        actions.append(add_task(
            UPGRADE_0_5_NODE, node_id=node_id, expected_class=expected_class, data=input_data,
            source_node_path=task_meta.get('source_node_path')
        ))
    else:
        return task_result(False, "Could not determine supported Open Badges object version for {}".format(node_id))

    return task_result(True, "Processed node {}".format(node_id), actions)


def _get_extension_actions(current_node, entry_path, new_contexts=None):
    new_actions = []
    if new_contexts is None:
        new_contexts = set()

    if not isinstance(current_node, dict):
        return new_actions

    if current_node.get('type'):
        types = list_of(current_node['type'])
        if 'Extension' in types:
            new_actions += [add_task(
                VALIDATE_EXTENSION_NODE,
                node_path=entry_path,
                node_json=json.dumps(current_node),
                context_urls=list(new_contexts),
                types_to_test=[t for t in types if t != 'Extension']
            )]

    for key in [k for k in current_node.keys() if k not in ('id', 'type',)]:
        val = current_node.get(key)
        if isinstance(val, list):
            for i in range(len(val)):
                new_actions += _get_extension_actions(val[i], entry_path + [key] + [i], new_contexts)
        else:
            new_actions += _get_extension_actions(val, entry_path + [key], new_contexts)

    return new_actions


def jsonld_compact_data(state, task_meta, **options):
    try:
        data = task_meta.get('data')
        if data and not sys.version[:3] < '3' and not isinstance(data, six.string_types):
            data = data.decode()
        input_data = json.loads(data)
        expected_class = task_meta.get('expected_class')
        depth = task_meta.get('depth')
    except TypeError:
        return task_result(False, "Could not load data")

    selected_options = options.get('jsonld_options', jsonld_use_cache)
    selected_options['documentLoader'].contexts = set()

    result = jsonld.compact(input_data, OPENBADGES_CONTEXT_V2_URI,
                            options=selected_options)

    new_contexts = list(selected_options['documentLoader'].contexts.copy())

    node_id = result.get('id', task_meta.get('node_id', get_next_blank_node_id()))

    # Handle mismatch between URL node source and declared ID.
    if result.get('id') and task_meta.get('node_id') and result['id'] != task_meta['node_id']:
        refetch_action = add_task(FETCH_HTTP_NODE, url=result['id'], depth=depth)
        if expected_class:
            refetch_action['expected_class'] = expected_class
        actions = [
            report_message(
                "Node fetched from source {} declared its id as {}".format(
                    task_meta['node_id'], node_id), MESSAGE_LEVEL_WARNING, success=False
            ),
            delete_outdated_node_tasks(task_meta['node_id']),
            refetch_action
        ]
        if task_meta.get('source_node_path'):
            actions.append(patch_node_reference(task_meta['source_node_path'], node_id))
        if state.get('report', {}).get('validationSubject') == task_meta['node_id']:
            actions.append(set_validation_subject(result['id']))

        return task_result(
            True,
            "Successfully compacted node {} from source {} and found ID mismatch".format(node_id, task_meta['node_id']),
            actions
        )

    actions = [
        add_node(node_id, data=result)
    ] + _get_extension_actions(result, [node_id], new_contexts)

    if expected_class:
        actions.append(
            add_task(VALIDATE_EXPECTED_NODE_CLASS, node_id=node_id,
                     expected_class=expected_class, depth=depth)
        )
    elif task_meta.get('detectAndValidateClass', True):
        actions.append(add_task(DETECT_AND_VALIDATE_NODE_CLASS, node_id=node_id, depth=depth))

    return task_result(
        True,
        "Successfully compacted node {}".format(node_id),
        actions
    )


def flatten_refetch_embedded_resource(state, task_meta, **options):
    try:
        node_id = task_meta['node_id']
        node = get_node_by_id(state, node_id)
        prop_name = task_meta['prop_name']
        node_class = task_meta['node_class']
        depth = task_meta['depth']
    except (IndexError, KeyError):
        raise TaskPrerequisitesError()

    actions = []
    value = node.get(prop_name)
    if value is None:
        return task_result(True, "Expected property {} was missing in node {}".format(prop_name, node_id))

    if isinstance(value, six.string_types) or (isinstance(value, list) and all(isinstance(s, six.string_types) for s in value)):
        return task_result(
            True, "Property {} referenced from {} is not embedded in need of flattening".format(
                prop_name, abv_node(node_id=node_id)
            ))

    if not isinstance(value, dict):
        if not isinstance(value, list):
            return task_result(
                False, "Property {} referenced from {} is not a JSON object, JSON object list, or string as expected".format(
                    prop_name, abv_node(node_id=node_id)
                ))
    else:
        value = [value]

    for node in value:
        embedded_node_id = node.get('id')

        if embedded_node_id is None:
            new_node = node.copy()
            embedded_node_id = '_:{}'.format(uuid.uuid4())
            new_node['id'] = embedded_node_id
            new_node['@context'] = OPENBADGES_CONTEXT_V2_URI
            actions.append(add_node(embedded_node_id, data=new_node))
            actions.append(patch_node(node_id, {prop_name: embedded_node_id}))
            actions.append(report_message(
                'Node id missing at {}. A blank node ID has been assigned'.format(
                    abv_node(node_path=[node_id, prop_name], length=64)
                ), message_level=MESSAGE_LEVEL_WARNING)
            )
        elif not isinstance(embedded_node_id, six.string_types) or not is_iri(embedded_node_id):
            return task_result(False, "Embedded JSON object at {} has no proper assigned id.".format(
                abv_node(node_path=[node_id, prop_name])))

        elif node_class == OBClasses.Assertion and not is_url(embedded_node_id):
            if not re.match(URN_REGEX, embedded_node_id, re.IGNORECASE):
                actions.append(report_message(
                    'ID format for {} at {} not in an expected HTTP or URN:UUID scheme'.format(
                        embedded_node_id, abv_node(node_path=[node_id, prop_name])
                    )))
            new_node = node.copy()
            new_node['@context'] = OPENBADGES_CONTEXT_V2_URI
            actions.append(add_node(embedded_node_id, data=node))
            actions.append(patch_node(node_id, {prop_name: embedded_node_id}))

        else:
            actions.append(patch_node(node_id, {prop_name: embedded_node_id}))
            if not node_match_exists(state, embedded_node_id) and not filter_tasks(
                state, node_id=embedded_node_id, task_type=FETCH_HTTP_NODE):
                # fetch
                actions.append(add_task(FETCH_HTTP_NODE, url=embedded_node_id, depth=depth))

    return task_result(True, "Embedded {} node in {} queued for storage and/or refetching as needed", actions)


from ..utils import MESSAGE_LEVEL_INFO, MESSAGE_LEVELS

from .utils import generate_task_key
from .action_types import (ADD_TASK, DELETE_OUTDATED_NODE_TASKS, DELETE_TASK, REPORT_MESSAGE, RESOLVE_TASK,
                           TRIGGER_CONDITION, UPDATE_TASK)


def add_task(task_name, **kwargs):
    # Ensure task is of a known type
    from ..tasks import task_types
    assert task_name in dir(task_types), '{} is not a known task'.format(task_name)

    task = {
        'type': ADD_TASK,
        'name': task_name,
        'task_key': kwargs.get('task_key', generate_task_key())
    }
    task.update(**kwargs)
    return task


def resolve_task(task_id, success=True, result=''):
    return {
        'type': RESOLVE_TASK,
        'task_id': task_id,
        'success': success,
        'result': result
    }


def trigger_condition(condition_key, result=''):
    return {
        'type': TRIGGER_CONDITION,
        'name': condition_key,
        'success': True,
        'result': result
    }


def delete_outdated_node_tasks(node_id):
    return {
        'type': DELETE_OUTDATED_NODE_TASKS,
        'node_id': node_id
    }


def delete_task(task_id):
    return {
        'type': DELETE_TASK,
        'task_id': task_id
    }


def update_task(task_id, task_name, **kwargs):
    # Ensure task is of a known type
    from ..tasks import task_types
    assert task_name in dir(task_types), '{} is not a known task'.format(task_name)

    task = {
        'type': UPDATE_TASK,
        'task_id': task_id,
        'name': task_name,
    }
    task.update(**kwargs)
    return task


def report_message(msg, message_level=MESSAGE_LEVEL_INFO, success=True):
    if message_level not in MESSAGE_LEVELS:
        raise TypeError("message")
    return {
        'type': REPORT_MESSAGE,
        'message': msg,
        'messageLevel': message_level,
        'success': success
    }



from future.standard_library import install_aliases
install_aliases()

import hashlib
import json
import string
import sys
try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

import requests
import requests_cache
from pyld.jsonld import JsonLdError


MESSAGE_LEVEL_ERROR = 'ERROR'
MESSAGE_LEVEL_WARNING = 'WARNING'
MESSAGE_LEVEL_INFO = 'INFO'
MESSAGE_LEVELS = (MESSAGE_LEVEL_ERROR, MESSAGE_LEVEL_WARNING, MESSAGE_LEVEL_INFO,)


class CachableDocumentLoader(object):
    def __init__(self, use_cache=False, backend='memory', expire_after=300, session=None):
        self.use_cache = use_cache
        self.contexts = set()

        if session is not None:
            self.session = session
        elif self.use_cache:
            self.session = requests_cache.CachedSession(backend=backend, expire_after=expire_after)
        else:
            self.session = requests.Session()

    def __call__(self, url):
        try:
            # validate URLs
            pieces = urlparse(url)
            if (not all([pieces.scheme, pieces.netloc]) or
                    pieces.scheme not in ['http', 'https'] or
                    set(pieces.netloc) > set(string.ascii_letters + string.digits + '-.:')):
                raise JsonLdError(
                    'Could not dereference URL; can only load URLs using',
                    'the "http" and "https" schemes.',
                    'jsonld.InvalidUrl', {'url': url},
                    code='loading document failed')

            response = self.session.get(
                url, headers={'Accept': 'application/ld+json, application/json'})

            doc = {'contextUrl': None, 'documentUrl': url, 'document': response.text}

            if self.use_cache:
                doc['from_cache'] = response.from_cache
                self.session.remove_expired_responses()

            # Save URL for Potential Extension contexts.
            try:
                data = json.loads(response.text)
                context = data['@context']
                if any([isinstance(el, dict) for el in list_of(context)]):
                    self.contexts.update([url])
            except Exception:
                pass

            return doc

        except JsonLdError as e:
            raise e
        except Exception as cause:
            raise JsonLdError(
                'Could not retrieve JSON-LD document from URL.',
                'jsonld.LoadDocumentError',
                code='loading document failed',
                cause=cause)


jsonld_use_cache = {'documentLoader': CachableDocumentLoader(use_cache=True)}
jsonld_no_cache = {'documentLoader': CachableDocumentLoader(use_cache=False)}


def list_of(value):
    if value is None:
        return []
    elif isinstance(value, list):
        return value
    return [value]


def identity_hash(identfier, salt='', alg='sha256'):

    if not sys.version[:3] < '3':
        identfier = identfier.encode()
        salt = salt.encode()
    if alg == 'sha256':
        return alg + '$' + hashlib.sha256(identfier + salt).hexdigest()
    elif alg == 'md5':
        return alg + '$' + hashlib.md5(identfier + salt).hexdigest()
    raise ValueError("Alg {} not supported.".format(alg))


def make_string_from_bytes(input_value):
    if isinstance(input_value, bytes):
        return input_value.decode()
    return input_value



import click
import json
import six

from .verifier import verify


@click.group()
def cli():
    pass


@click.command(name='verify')
@click.argument(u'input_file', type=click.File('rb'), required=False)
@click.argument(u'output_file', type=click.File('wb'), required=False)
@click.option(u'--data', type=str, default=None,
              help=u'Open Badges URL, JSON, or JWS-signed string input to verify')
@click.option(u'--recipient', type=str, prompt=False,
              help=u'Open Badges Profile JSON trusted to describe to the recipient')
def verify_badge_input(input_file, output_file, data, recipient):
    """
    This command takes Open Badges input in several formats and returns validation results.

    Positional Arguments:

    \b
      Input filename:    File must exist.
    \b
      Output filename:   If file exists, it will be overwritten.
    """
    if recipient is not None:
        try:
            recipient = json.loads(recipient)
        except Exception:
            raise click.Abort("Could not interpret profile input (expected JSON).")

    if data is not None and not isinstance(data, six.string_types):
        raise click.Abort("Expected data to be input as a string")

    results = verify(data, recipient_profile=recipient)
    is_valid = "Badge input is valid." if results['report'].get('valid') else "Badge input is not valid."

    click.echo(json.dumps(results, indent=4), file=output_file)
    click.echo(is_valid)


cli.add_command(verify_badge_input)


if __name__ == '__main__':
    cli()



import json
from pyld import jsonld
from requests_cache import CachedSession
import responses
import unittest

from openbadges.verifier.utils import CachableDocumentLoader

try:
    from .testfiles.test_components import test_components
    from tests.utils import set_up_context_mock
except (ImportError, SystemError):
    from .testfiles.test_components import test_components
    from .testutils import set_up_context_mock


class DocumentLoaderTests(unittest.TestCase):
    @responses.activate
    def test_that_caching_loader_uses_local_cache(self):
        url = 'http://example.com/assertionmaybe'
        loadurl = CachableDocumentLoader(use_cache=True)
        data = test_components['2_0_basic_assertion']
        responses.add(
            responses.GET, url, body=data, status=200, content_type='application/ld+json')

        first_remote_document = loadurl(url)
        second_remote_document = loadurl(url)

        self.assertFalse(first_remote_document.get('from_cache'))
        self.assertTrue(second_remote_document.get('from_cache'))
        self.assertEqual(
            first_remote_document.get('document'), second_remote_document.get('document'))
        self.assertEqual(first_remote_document.get('document'), data)

    @responses.activate
    def test_that_noncaching_loader_loads_url(self):
        url = 'http://example.com/assertionmaybe'
        loadurl = CachableDocumentLoader(use_cache=False)
        data = test_components['2_0_basic_assertion']
        responses.add(
            responses.GET, url, body=data, status=200, content_type='application/ld+json')

        document = loadurl(url)

        self.assertEqual(document.get('from_cache', 'Uncached'), 'Uncached')
        self.assertEqual(document.get('document'), data)

    @responses.activate
    def test_that_pyld_accepts_caching_loader_for_compaction(self):
        assertion_data = json.loads(test_components['2_0_basic_assertion'])
        context_url = assertion_data['@context']
        loadurl = CachableDocumentLoader(use_cache=True)
        set_up_context_mock()

        first_compacted = jsonld.compact(
            assertion_data, context_url, options={'documentLoader': loadurl})
        second_compacted = jsonld.compact(
            assertion_data, context_url, options={'documentLoader': loadurl})
        # in order to have 'HostedBadge' as the verification type, the assertion_data
        # needs to have gone through compaction against the openbadges context document
        self.assertEqual(first_compacted['verification']['type'], 'HostedBadge')
        # second compaction should have built from the cache
        self.assertEqual(first_compacted['verification']['type'],
                         second_compacted['verification']['type'])

    @responses.activate
    def test_basic_response_config(self):
        """
        Here is how to load something into a session cache with requests_cache and responses
        """
        session = CachedSession(backend='memory', expire_after=100000)

        url = 'http://example.com/123'
        data = {'foo': 'yes please'}
        responses.add(responses.GET, url, json=data)
        response_one = session.get(url)
        self.assertFalse(response_one.from_cache)
        responses.reset()
        responses.add(responses.GET, url, json={})

        response_two = session.get(url)
        self.assertTrue(response_two.from_cache)
        self.assertEqual(response_two.json()['foo'], data['foo'])

    def test_loader_with_session(self):
        session = CachedSession(backend='memory', expire_after=100000)
        loadurl = CachableDocumentLoader(use_cache=True, session=session)
        assertion_data = json.loads(test_components['2_0_basic_assertion'])
        context_url = assertion_data['@context']

        set_up_context_mock()
        session.get(context_url)  # precache response
        responses.reset()
        responses.add(responses.GET, context_url, json={'nothing': 'happenin'})

        first_compacted = jsonld.compact(
            assertion_data, context_url, options={'documentLoader': loadurl})
        second_compacted = jsonld.compact(
            assertion_data, context_url, options={'documentLoader': loadurl})

        # second compaction should have built from the cache
        self.assertEqual(first_compacted['verification']['type'],
                         second_compacted['verification']['type'])



import datetime
import json
import responses
import unittest
import sys

from openbadges.verifier.actions.graph import add_node
from openbadges.verifier.actions.tasks import add_task
from openbadges.verifier.extensions import GeoLocation, ExampleExtension, ApplyLink
from openbadges.verifier.openbadges_context import OPENBADGES_CONTEXT_V2_URI
from openbadges.verifier.reducers import main_reducer
from openbadges.verifier.reducers.graph import graph_reducer
from openbadges.verifier.state import INITIAL_STATE
from openbadges.verifier.tasks.extensions import validate_extension_node
from openbadges.verifier.tasks.graph import _get_extension_actions
from openbadges.verifier.tasks import task_named
from openbadges.verifier.tasks.task_types import (INTAKE_JSON, JSONLD_COMPACT_DATA, VALIDATE_EXTENSION_NODE,
                                         VALIDATE_EXTENSION_SINGLE)
from openbadges.verifier.tasks.utils import combine_contexts
from openbadges.verifier.utils import jsonld_no_cache, CachableDocumentLoader
from openbadges.verifier.verifier import extension_validation_store, generate_report

from tests.utils import set_up_context_mock


class CompactJsonExtensionDiscoveryTests(unittest.TestCase):
    def test_can_discover_actions_right(self):
        node = {
            'string_prop': 'string_val'
        }
        self.assertEqual(_get_extension_actions(node, ['_:b0']), [])

        node['dict_prop_1'] = {'type': 'Extension'}
        actions = _get_extension_actions(node, ['_:b0'])
        self.assertEqual(len(actions), 1,
                         "When one Extension-type node is present, file one action")
        self.assertEqual(actions[0]['node_path'], ['_:b0', 'dict_prop_1'])

        node['dict_prop_1'] = {'type': ['Extension', 'extensions:ApplyLink']}
        actions = _get_extension_actions(node, ['_:b0'])
        self.assertEqual(len(actions), 1,
                         "It can handle an Extension-type node declared in list")
        self.assertEqual(actions[0]['node_path'], ['_:b0', 'dict_prop_1'])

        node['dict_prop_2'] = {'type': 'NotAnExtension'}
        self.assertEqual(len(_get_extension_actions(node, ['_:b0'])), 1,
                         "Another non-Extension node doesn't add another action.")

        node['dict_prop_2'] = {'type': 'Extension'}
        self.assertEqual(len(_get_extension_actions(node, ['_:b0'])), 2,
                         "A second Extension node yields another action.")

        node = {
            'dict_prop_3': {
                'string_prop_2': 'string_val',
                'dict_prop_4': {'type': 'Extension'}
            }
        }
        actions = _get_extension_actions(node, ['_:b0'])
        self.assertEqual(len(actions), 1, "One Extension is found.")
        self.assertEqual(actions[0]['node_path'], ['_:b0', 'dict_prop_3', 'dict_prop_4'],
                         "A deeply nested extension is properly identified.")

        node = {
            'list_prop_1': [
                {
                    'prop': 'not an extension'
                },
                {
                    'id': '_:b0',
                    'string_prop_1': 'string_val',
                    'dict_prop_1': {'id': '_:b1'}
                },
                'string_val'
            ]
        }

        actions = _get_extension_actions(node, ['_:b0'])
        self.assertEqual(len(actions), 0, "No extensions exist in node yet")
        node['list_prop_1'][1]['type'] = 'Extension'
        actions = _get_extension_actions(node, ['_:b0'])
        self.assertEqual(len(actions), 1, "An Extension is found inside a many=True value.")
        self.assertEqual(
            actions[0]['node_path'], ['_:b0', 'list_prop_1', 1],
            "The action's node_path correctly identifies the list index of the Extension")


class ExtensionNodeValidationTests(unittest.TestCase):
    def setUp(self):
        self.extension = {
            'id': '_:b0',
            'type': ['Extension', 'extensions:ExampleExtension'],
            'http://schema.org/text': 'I\'m a property, short and sweet'
        }
        self.evidence = {
            'id': '_:b1',
            'narrative': 'Rocked the free world'
        }
        self.first_node = {
            '@context': [OPENBADGES_CONTEXT_V2_URI, ExampleExtension.context_url],
            'id': 'http://example.org/assertion',
            'extensions:exampleExtension': self.extension,
            'evidence': self.evidence
        }

    def load_mocks(self):
        loader = CachableDocumentLoader(use_cache=True)
        loader.session.cache.remove_old_entries(datetime.datetime.utcnow())
        loader.contexts = set()
        self.options = {'jsonld_options': {'documentLoader': loader}}

        set_up_context_mock()
        loader(OPENBADGES_CONTEXT_V2_URI)
        schema_url = list(ExampleExtension.validation_schema)[0]
        responses.add(responses.GET, ExampleExtension.context_url, status=200, json=ExampleExtension.context_json)
        loader(ExampleExtension.context_url)
        responses.add(responses.GET, schema_url, status=200, json=ExampleExtension.validation_schema[schema_url])
        loader.session.get(schema_url)

        self.state = INITIAL_STATE
        task = add_task(
            INTAKE_JSON, data=json.dumps(self.first_node), node_id=self.first_node['id'])
        result, message, actions = task_named(INTAKE_JSON)(self.state, task,  **self.options)
        self.state = main_reducer(self.state, actions[0])
        result, message, actions = task_named(actions[1]['name'])(
            self.state, actions[1],  **self.options)  # JSONLD_COMPACT_DATE
        self.state = main_reducer(self.state, actions[0])  # ADD_NODE
        self.validation_task = actions[1]  # VALIDATE_EXTENSION_NODE


    @responses.activate
    def test_validate_extension_node_basic(self):
        self.load_mocks()
        task_meta = self.validation_task

        result, message, actions = validate_extension_node(self.state, task_meta, **self.options)
        self.assertTrue(result, "A valid expression of the extension should pass")
        self.assertIn('validated on node', message)
        self.assertEqual(len(actions), 0)

    @responses.activate
    def test_validate_extension_node_invalid(self):
        self.load_mocks()
        task_meta = self.validation_task

        # String value is required, we'll try a number
        self.state['graph'][0]['extensions:exampleExtension']['schema:text'] = 1337

        result, message, actions = validate_extension_node(self.state, task_meta, **self.options)
        self.assertFalse(result, "An invalid expression of a rule in schema should fail")
        self.assertIn('did not validate', message)
        self.assertEqual(len(actions), 0)

    @responses.activate
    def test_validation_breaks_down_multiple_extensions(self):
        self.load_mocks()
        # Load up ApplyLink schema and context
        responses.add(responses.GET, ApplyLink.context_url, status=200, json=ApplyLink.context_json)
        self.options['jsonld_options']['documentLoader'](ApplyLink.context_url)
        schema_url = list(ApplyLink.validation_schema)[0]
        responses.add(responses.GET, schema_url, status=200, json=ApplyLink.validation_schema[schema_url])
        self.options['jsonld_options']['documentLoader'].session.get(schema_url)

        self.state['graph'][0]['extensions:exampleExtension']['type'].append('extensions:ApplyLink')
        task_meta = self.validation_task.copy()
        task_meta['context_urls'].append(ApplyLink.context_url)
        task_meta['types_to_test'].append('extensions:ApplyLink')

        result, message, actions = validate_extension_node(self.state, task_meta, **self.options)
        self.assertTrue(result, "Task breakdown should succeed.")
        self.assertIn('Multiple extension types', message)
        self.assertEqual(len(actions), 2)
        self.assertTrue(all(a['name'] == VALIDATE_EXTENSION_SINGLE for a in actions),
                        'All tasks created should be of correct type')

        for action in actions:
            aresult, amessage, aactions = validate_extension_node(self.state, task_meta, **self.options)
            self.assertTrue(aresult)


class ComplexExtensionNodeValdiationTests(unittest.TestCase):
    """
    Tests for extensions that use nested properties.
    """
    @responses.activate
    def test_node_json_validation(self):
        node = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'id': 'http://example.com/1',
            'type': 'Assertion',
            'schema:location': {
                '@context': 'https://w3id.org/openbadges/extensions/geoCoordinatesExtension/context.json',
                'type': ['Extension', 'extensions:GeoCoordinates'],
                'description': 'That place in the woods where we built the fort',
                'schema:geo': {
                    'schema:latitude': 44.580900,
                    'schema:longitude': -123.301815
                }
            }
        }

        loader = CachableDocumentLoader(use_cache=True)
        loader.session.cache.remove_old_entries(datetime.datetime.utcnow())
        loader.contexts = set()
        options = {'jsonld_options': {'documentLoader': loader}}

        set_up_context_mock()
        loader(OPENBADGES_CONTEXT_V2_URI)
        schema_url = list(GeoLocation.validation_schema)[0]
        responses.add(responses.GET, GeoLocation.context_url, status=200, json=GeoLocation.context_json)
        loader(GeoLocation.context_url)
        responses.add(responses.GET, schema_url, status=200, json=GeoLocation.validation_schema[schema_url])
        loader.session.get(schema_url)

        state = INITIAL_STATE
        task = add_task(
            INTAKE_JSON, data=json.dumps(node), node_id=node['id'])
        result, message, actions = task_named(INTAKE_JSON)(state, task, **options)
        state = main_reducer(state, actions[0])
        result, message, actions = task_named(actions[1]['name'])(
            state, actions[1], **options)  # JSONLD_COMPACT_DATE
        state = main_reducer(state, actions[0])  # ADD_NODE
        task_meta = actions[1]  # VALIDATE_EXTENSION_NODE

        result, message, actions = validate_extension_node(state, task_meta)
        self.assertTrue(result, "A valid expression of the extension should pass")
        self.assertIn('validated on node', message)
        self.assertEqual(len(actions), 0)

        del state['graph'][0]['schema:location']['schema:geo']['schema:latitude']
        result, message, actions = validate_extension_node(state, task_meta)
        self.assertFalse(result, "A required property not present should be detected by JSON-schema.")

    @responses.activate
    def test_extension_discovered_jsonld_compact(self):
        """
        Ensure an extension node is properly discovered and that the task runs without error.
        """
        node = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'id': 'http://example.com/1',
            'type': 'Assertion',
            'schema:location': {
                '@context': GeoLocation.context_url,
                'type': ['Extension', 'extensions:GeoCoordinates'],
                'description': 'That place in the woods where we built the fort',
                'schema:geo': {
                    'schema:latitude': 44.580900,
                    'schema:longitude': -123.301815
                }
            }
        }
        state = INITIAL_STATE

        set_up_context_mock()

        responses.add(
            responses.GET,
            GeoLocation.context_url,
            body=json.dumps(GeoLocation.context_json),
            status=200,
            content_type='application/ld+json')

        schema_url = 'https://w3id.org/openbadges/extensions/geoCoordinatesExtension/schema.json'
        responses.add(
            responses.GET, schema_url,
            body=json.dumps(GeoLocation.validation_schema[schema_url]),
            status=200,
            content_type='application/ld+json')

        compact_task = add_task(
            JSONLD_COMPACT_DATA, data=json.dumps(node), jsonld_options=jsonld_no_cache,
            context_urls=[GeoLocation.context_url]
        )
        result, message, actions = task_named(JSONLD_COMPACT_DATA)(state, compact_task)
        self.assertTrue(result, "JSON-LD Compact is successful.")
        self.assertIn(VALIDATE_EXTENSION_NODE, [i.get('name') for i in actions], "Validation task queued.")
        state = main_reducer(state, actions[0])  # ADD_NODE

        validate_task = [i for i in actions if i.get('name') == VALIDATE_EXTENSION_NODE][0]
        self.assertIsNotNone(validate_task['node_json'])

        result, message, actions = task_named(VALIDATE_EXTENSION_NODE)(state, validate_task)
        self.assertTrue(result, "Validation task is successful.")

    @responses.activate
    def test_extension_discovered_jsonld_compact_with_validates_type_variation(self):
        """
        Ensure an extension node is properly discovered and that the task runs without error.
        """
        node = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'id': 'http://example.com/1',
            'type': 'Assertion',
            'schema:location': {
                '@context': GeoLocation.context_url,
                'type': ['Extension', 'extensions:GeoCoordinates'],
                'description': 'That place in the woods where we built the fort',
                'schema:geo': {
                    'schema:latitude': 44.580900,
                    'schema:longitude': -123.301815
                }
            }
        }
        state = INITIAL_STATE

        set_up_context_mock()
        geo_context = GeoLocation.context_json
        geo_context["obi:validation"][0]['obi:validatesType'] = 'https://w3id.org/openbadges/extensions#GeoCoordinates'

        responses.add(
            responses.GET,
            GeoLocation.context_url,
            body=json.dumps(geo_context),
            status=200,
            content_type='application/ld+json')

        schema_url = 'https://w3id.org/openbadges/extensions/geoCoordinatesExtension/schema.json'
        responses.add(
            responses.GET, schema_url,
            body=json.dumps(GeoLocation.validation_schema[schema_url]),
            status=200,
            content_type='application/ld+json')

        compact_task = add_task(
            JSONLD_COMPACT_DATA, data=json.dumps(node), jsonld_options=jsonld_no_cache,
            context_urls=[GeoLocation.context_url]
        )
        result, message, actions = task_named(JSONLD_COMPACT_DATA)(state, compact_task)
        self.assertTrue(result, "JSON-LD Compact is successful.")
        self.assertIn(VALIDATE_EXTENSION_NODE, [i.get('name') for i in actions], "Validation task queued.")
        state = main_reducer(state, actions[0])  # ADD_NODE

        validate_task = [i for i in actions if i.get('name') == VALIDATE_EXTENSION_NODE][0]
        self.assertIsNotNone(validate_task['node_json'])

        result, message, actions = task_named(VALIDATE_EXTENSION_NODE)(state, validate_task)
        self.assertTrue(result, "Validation task is successful.")

    @responses.activate
    def test_extension_discovered_jsonld_compact_with_context_variation(self):
        """
        Ensure an extension node is properly discovered and that the task runs without error.
        """
        node = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'id': 'http://example.com/1',
            'type': 'Assertion',
            'schema:location': {
                '@context': GeoLocation.context_url,
                'type': ['Extension', 'extensions:GeoCoordinates'],
                'description': 'That place in the woods where we built the fort',
                'schema:geo': {
                    'schema:latitude': 44.580900,
                    'schema:longitude': -123.301815
                }
            }
        }
        state = INITIAL_STATE

        set_up_context_mock()
        geo_context = GeoLocation.context_json
        geo_context['@context'] = [OPENBADGES_CONTEXT_V2_URI, geo_context['@context'].copy()]

        responses.add(
            responses.GET,
            GeoLocation.context_url,
            body=json.dumps(geo_context),
            status=200,
            content_type='application/ld+json')

        schema_url = 'https://w3id.org/openbadges/extensions/geoCoordinatesExtension/schema.json'
        responses.add(
            responses.GET, schema_url,
            body=json.dumps(GeoLocation.validation_schema[schema_url]),
            status=200,
            content_type='application/ld+json')

        compact_task = add_task(
            JSONLD_COMPACT_DATA, data=json.dumps(node), jsonld_options=jsonld_no_cache,
            context_urls=[GeoLocation.context_url]
        )
        result, message, actions = task_named(JSONLD_COMPACT_DATA)(state, compact_task)
        self.assertTrue(result, "JSON-LD Compact is successful.")
        self.assertIn(VALIDATE_EXTENSION_NODE, [i.get('name') for i in actions], "Validation task queued.")
        state = main_reducer(state, actions[0])  # ADD_NODE

        validate_task = [i for i in actions if i.get('name') == VALIDATE_EXTENSION_NODE][0]
        self.assertIsNotNone(validate_task['node_json'])

        result, message, actions = task_named(VALIDATE_EXTENSION_NODE)(state, validate_task)
        self.assertTrue(result, "Validation task is successful.")


class UnknownExtensionsTests(unittest.TestCase):
    """
    TODO: In the future, dynamic discovery of extensions will be possible.
    Until then, make sure we are reporting on unverified extensions.
    """
    def test_report_message_on_unknown_extension(self):
        first_node = {
            'id': 'http://example.org/assertion',
            'extensions:exampleExtension': '_:b0',
            'evidence': '_:b1'
        }
        extension = {
            'id': '_:b0',
            'type': ['Extension', 'extensions:UnknownExtension'],
            'schema:unknownProperty': 'I\'m a property, short and sweet'
        }
        state = {'graph': [first_node, extension]}
        task_meta = add_task(
            VALIDATE_EXTENSION_NODE, node_id=extension['id'])

        result, message, actions = validate_extension_node(state, task_meta)
        self.assertFalse(result, "An unknown extension will fail for now.")


class DynamicExtensionValidationTests(unittest.TestCase):
    """
    Extension validation involves establishing mocks for the context & schema resources
    """
    def set_up_test_extension(self):
        self.extension_schema = {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "1.1 Open Badge Example Extension for testing: Unknown Extension",
            "description": "An extension that allows you to add a single string unknownProperty to an extension object for unknown reasons.",
            "type": "object",
            "properties": {
                "unknownProperty": {
                    "type": "string"
                }
            },
            "required": ["unknownProperty"]
        }
        self.extension_schema_url = 'http://example.org/unkownSchema'
        self.extension_context = {
            '@context': {
                "obi": "https://w3id.org/openbadges#",
                "extensions": "https://w3id.org/openbadges/extensions#",
                'unknownProperty': 'http://schema.org/unknownProperty'
            },
            "obi:validation": [
                {
                    "obi:validatesType": "extensions:UnknownExtension",
                    "obi:validationSchema": self.extension_schema_url
                }
            ]
        }
        self.extension_context_url = 'http://example.org/unknownExtensionContext'

        responses.add(
            responses.GET, self.extension_context_url,
            json=self.extension_context
        )
        responses.add(
            responses.GET, self.extension_schema_url,
            json=self.extension_schema
        )

    @responses.activate
    def test_queue_validation_on_unknown_extension(self):
        set_up_context_mock()
        self.set_up_test_extension()

        first_node_json = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'id': 'http://example.org/assertion',
            'extensions:exampleExtension': {
                '@context': self.extension_context_url,
                'type': ['Extension', 'extensions:UnknownExtension'],
                'unknownProperty': 'I\'m a property, short and sweet'
            },
            'evidence': 'http://example.org/evidence'
        }

        state = INITIAL_STATE

        task_meta = add_task(
            INTAKE_JSON, data=json.dumps(first_node_json), node_id=first_node_json['id'])

        result, message, actions = task_named(INTAKE_JSON)(state, task_meta)
        for action in actions:
            state = main_reducer(state, action)

        # Compact JSON
        result, message, actions = task_named(state['tasks'][0]['name'])(state, state['tasks'][0])

        self.assertEqual(len(actions), 3)

        state = main_reducer(state, actions[0])

        validation_action = actions[1]
        result, message, actions = validate_extension_node(state, validation_action)

        self.assertTrue(result)

    @responses.activate
    def validate_basic_standalone_extension_node(self):
        set_up_context_mock()
        self.set_up_test_extension()

        extension_data = {
            '@context': self.extension_context_url,
            '@type': ['Extension', 'extensions:UnknownExtension'],
            'unknownProperty': 'I\'m a property, short and sweet'
        }

        store = extension_validation_store(extension_data)
        report = generate_report(store)

        self.assertTrue(report['report']['valid'])

        extension_data['unknownProperty'] = 42  # provokes extension schema error: should be a string.

        store = extension_validation_store(extension_data)
        report = generate_report(store)

        self.assertFalse(report['report']['valid'])

        del extension_data['@type']
        extension_data['unknownProperty'] = "Ok, valid string again"

        store = extension_validation_store(extension_data)
        report = generate_report(store)

        self.assertFalse(
            report['report']['valid'],
            "Should report an error if there weren't any discoverable extension types to test."
        )


class ExtensionContextCombinations(unittest.TestCase):
    def test_combine_contexts(self):
        str1, str2 = '_:str1', '_:str2'
        dict1, dict2 = {'dict1': 'http://dict1.com'}, {'dict2': 'http://dict2.com'}
        dict3 = {'@context': [str2, dict1]}
        list1, list2 = ['_:list1', '_:list1b'], ['_:list2']

        self.assertEqual(combine_contexts(str1, str2),          [str1, str2])
        self.assertEqual(combine_contexts(str1, dict1),         [str1, dict1])
        self.assertEqual(combine_contexts(list1, list2),        list1 + list2)
        self.assertEqual(combine_contexts(list1, dict1, str1),  list1 + [dict1, str1])
        self.assertEqual(combine_contexts(list1, dict3, list2), list1 + [str2, dict1] + list2)



import json
import unittest

try:
    from .testfiles.test_components import test_components
except (ImportError, SystemError):
    from .testfiles.test_components import test_components


class TestLoadComponents(unittest.TestCase):
    def test_load_assertion_from_test_components(self):
        assertion = test_components['2_0_basic_assertion']
        assertion_value = json.loads(assertion)
        self.assertEqual(assertion_value['issuedOn'], "2016-12-31T23:59:59Z")



import unittest

from openbadges.verifier.actions.tasks import add_task
from openbadges.verifier.tasks import run_task, task_named
from openbadges.verifier.tasks.task_types import DETECT_AND_VALIDATE_NODE_CLASS


class RelatedObjectTests(unittest.TestCase):
    def test_validate_related_language(self):
        options = {'max_validation_depth': 3}

        assertion = {
            'type': 'Assertion',
            'id': 'http://example.com/assertion',
            'verification': {
                'type': 'HostedBadge'
            },
            'badge': 'http://example.com/badgeclass'
        }
        badgeclass = {
            'id': 'http://example.com/badgeclass',
            'type': 'BadgeClass',
            '@language': 'es',
            'issuer': 'http://example.com/issuer',
            'related': {
                'id': 'http://example.com/other_badgeclass',
                '@language': 'en-US'
            },
            'name': 'Insignia Pronto'
        }
        state = {'graph': [assertion, badgeclass]}
        task_meta = add_task(DETECT_AND_VALIDATE_NODE_CLASS, node_id=badgeclass['id'], depth=0)
        result, message, actions = task_named(DETECT_AND_VALIDATE_NODE_CLASS)(state, task_meta, **options)
        self.assertTrue(result)

        language_task = [t for t in actions if t.get('prop_name') == '@language'][0]
        r, _, __ = run_task(state, language_task)
        self.assertTrue(r, "The BadgeClass's language property is valid.")

        related_task = [t for t in actions if t.get('prop_name') == 'related'][0]
        result, message, actions = task_named(related_task['name'])(state, related_task, **options)
        self.assertTrue(result, "The related property is valid and queues up task discovery for embedded node")
        result, message, actions = task_named(actions[0]['name'])(state, actions[0], **options)
        self.assertTrue(result, "Some tasks are discovered to validate the related node.")
        self.assertEqual(len(actions), 2, "There are only tasks for 'id' and '@language'.")

        for a in actions:
            r, _, __ = run_task(state, a)
            self.assertTrue(r, "Related node property validation is successful.")

    def test_validate_multiple_related_languages(self):
        options = {'max_validation_depth': 3}

        assertion = {
            'type': 'Assertion',
            'id': 'http://example.com/assertion',
            'verification': {
                'type': 'HostedBadge'
            },
            'badge': 'http://example.com/badgeclass'
        }
        badgeclass = {
            'id': 'http://example.com/badgeclass',
            'type': 'BadgeClass',
            '@language': 'es',
            'issuer': 'http://example.com/issuer',
            'related': [{
                'id': 'http://example.com/other_badgeclass',
                '@language': 'en-US'
            },
            {
                'id': 'http://example.com/another_badgeclass',
                '@language': 'fi'
            }],
            'name': 'Insignia Pronto'
        }
        state = {'graph': [assertion, badgeclass]}
        task_meta = add_task(DETECT_AND_VALIDATE_NODE_CLASS, node_id=badgeclass['id'], depth=0)
        result, message, actions = task_named(DETECT_AND_VALIDATE_NODE_CLASS)(state, task_meta, **options)
        self.assertTrue(result)

        language_task = [t for t in actions if t.get('prop_name') == '@language'][0]
        r, _, __ = task_named(language_task['name'])(state, language_task, **options)
        self.assertTrue(r, "The BadgeClass's language property is valid.")

        related_task = [t for t in actions if t.get('prop_name') == 'related'][0]
        result, message, actions = task_named(related_task['name'])(state, related_task, **options)
        self.assertTrue(result, "The related property is valid and queues up task discovery for embedded node")
        self.assertEqual(len(actions), 2, "It has now discovered two nodes to test.")



from base64 import b64encode
from Crypto.PublicKey import RSA
from jose import jws
import json
import responses
import unittest
import sys

from openbadges.verifier.actions.graph import patch_node
from openbadges.verifier.actions.tasks import add_task
from openbadges.verifier.exceptions import TaskPrerequisitesError
from openbadges.verifier.openbadges_context import OPENBADGES_CONTEXT_V2_URI
from openbadges.verifier.reducers.graph import graph_reducer
from openbadges.verifier.tasks.crypto import (process_jws_input, verify_key_ownership, verify_jws_signature,
                                     verify_signed_assertion_not_revoked,)
from openbadges.verifier.tasks.task_types import (PROCESS_JWS_INPUT, VERIFY_JWS, VERIFY_KEY_OWNERSHIP,
                                         VERIFY_SIGNED_ASSERTION_NOT_REVOKED,)
from openbadges.verifier.verifier import verify
from openbadges.verifier.utils import make_string_from_bytes

try:
    from .testfiles.test_components import test_components
    from tests.utils import set_up_context_mock, set_up_image_mock
except (ImportError, SystemError):
    from .testfiles.test_components import test_components
    from .utils import set_up_context_mock, set_up_image_mock


class JwsVerificationTests(unittest.TestCase):
    def setUp(self):
        self.key = RSA.generate(2048)
        self.public_key_pem = self.key.publickey().export_key()
        self.signing_key_doc = {
            'id': 'http://example.org/key1',
            'type': 'CryptographicKey',
            'owner': 'http://example.org/issuer',
            'publicKeyPem': self.public_key_pem
        }
        self.issuer_data = {
            'id': 'http://example.org/issuer',
            'publicKey': 'http://example.org/key1'
        }
        self.badgeclass = {
            'id': '_:b1',
            'issuer': 'http://example.org/issuer'
        }
        self.verification_object = {
            'id': '_:b0',
            'type': 'SignedBadge',
            'creator': 'http://example.org/key1'
        }
        self.assertion_data = {
            'id': 'urn:uuid:bf8d3c3d-fe60-487c-87a3-06440d0d0163',
            'verification': '_:b0',
            'badge': '_:b1'
        }
        self.assertion = {
            'id': 'urn:uuid:bf8d3c3d-fe60-487c-87a3-06440d0d0163',
            'verification': self.verification_object,
            'badge': self.badgeclass
        }

        self.signature = jws.sign(self.assertion, self.key, algorithm='RS256')

        self.state = {
            'graph': [self.signing_key_doc, self.issuer_data, self.badgeclass,
                      self.verification_object, self.assertion_data]
        }

    def test_can_process_jws_input(self):
        task_meta = add_task(PROCESS_JWS_INPUT, data=self.signature, depth=0)
        state = {}

        success, message, actions = process_jws_input(state, task_meta)
        self.assertTrue(success)
        self.assertEqual(len(actions), 3)

    def test_can_verify_jws(self):
        task_meta = add_task(VERIFY_JWS, data=self.signature,
                             node_id=self.assertion_data['id'], depth=0)

        success, message, actions = verify_jws_signature(self.state, task_meta)
        self.assertTrue(success)
        self.assertEqual(len(actions), 2)

        # Construct an invalid signature by adding to payload after signing, one theoretical attack.
        signed = jws.sign(self.assertion, self.key, algorithm='RS256')
        self.assertion_data['evidence'] = 'http://hahafakeinserteddata.com'

        encoded_separator = '.'
        if sys.version[:3] < '3':
            original_header, original_payload, original_signature = self.signature.split(encoded_separator)
            encoded_payload = b64encode(json.dumps(self.assertion_data))
        else:
            encoded_separator = '.'.encode()
            original_header, original_payload, original_signature = self.signature.encode().split(encoded_separator)
            encoded_payload = b64encode(json.dumps(self.assertion_data).encode())

        self.signed_assertion = encoded_separator.join((original_header, encoded_payload, original_signature))

        task_meta = add_task(VERIFY_JWS, data=self.signed_assertion,
                             node_id=self.assertion_data['id'], depth=0)

        success, message, actions = verify_jws_signature(self.state, task_meta)
        self.assertFalse(success)
        self.assertEqual(len(actions), 2)

    def test_can_verify_key_ownership(self):
        state = self.state
        task_meta = add_task(VERIFY_KEY_OWNERSHIP, node_id=self.assertion_data['id'], depth=0)

        result, message, actions = verify_key_ownership(state, task_meta)
        self.assertTrue(result)

        del self.verification_object['creator']
        with self.assertRaises(TaskPrerequisitesError):
            verify_key_ownership(state, task_meta)

        self.verification_object['creator'] = 'http://nowhere.man'
        with self.assertRaises(TaskPrerequisitesError):
            verify_key_ownership(state, task_meta)

        self.verification_object['creator'] = self.signing_key_doc['id']
        self.issuer_data['publicKey'] = ['http://example.org/key2']
        result, message, actions = verify_key_ownership(state, task_meta)
        self.assertFalse(result)
        self.issuer_data['publicKey'] = [self.signing_key_doc['id'], 'http://example.org/key2']
        result, message, actions = verify_key_ownership(state, task_meta)
        self.assertTrue(result)
        self.assertEqual(len(actions), 0)

        self.issuer_data['revocationList'] = 'http://example.org/revocationList'
        result, message, actions = verify_key_ownership(state, task_meta)
        self.assertTrue(result)
        self.assertEqual(len(actions), 1, "Revocation check task should be queued.")
        self.assertTrue(actions[0]['name'], VERIFY_SIGNED_ASSERTION_NOT_REVOKED)

    def test_can_verify_revoked(self):
        state = self.state
        revocation_list = {
            'id': 'http://example.org/revocationList',
            'type': 'RevocationList',
            'revokedAssertions': []
        }
        state['graph'] += [revocation_list]
        self.issuer_data['revocationList'] = revocation_list['id']

        task_meta = add_task(VERIFY_SIGNED_ASSERTION_NOT_REVOKED, node_id=self.assertion_data['id'])

        result, message, actions = verify_signed_assertion_not_revoked(state, task_meta)
        self.assertTrue(result)

        b123 = {'id': 'http://example.org/another', 'revocationReason': 'was imaginary'}
        revocation_list['revokedAssertions'] = [
            self.assertion_data['id'], 'http://example.org/else',
            'http://example.org/another'
        ]
        state['graph'].append(b123)
        result, message, actions = verify_signed_assertion_not_revoked(state, task_meta)
        self.assertFalse(result)

        revocation_entry = {
            'id': self.assertion_data['id'],
            'revocationReason': 'Tom got to pressing the award button again. Oh, Tom.'}
        revocation_list['revokedAssertions'].append(revocation_entry)
        result, message, actions = verify_signed_assertion_not_revoked(state, task_meta)
        self.assertFalse(result)
        self.assertIn(revocation_entry['revocationReason'], message)


class JwsFullVerifyTests(unittest.TestCase):
    @responses.activate
    def test_can_full_verify_jws_signed_assertion(self):
        """
        I can input a JWS string
        I can extract the Assertion from the input signature string and store it as the canonical version of the Assertion.
        I can discover and retrieve key information from the Assertion.
        I can Access the signing key
        I can verify the key is associated with the listed issuer Profile
        I can verify the JWS signature has been created by a key trusted to correspond to the issuer Profile
        Next: I can verify an assertion with an ephemeral embedded badgeclass as well
        """
        input_assertion = json.loads(test_components['2_0_basic_assertion'])
        input_assertion['verification'] = {'type': 'signed', 'creator': 'http://example.org/key1'}
        set_up_image_mock(u'https://example.org/beths-robot-badge.png')

        input_badgeclass = json.loads(test_components['2_0_basic_badgeclass'])
        set_up_image_mock(input_badgeclass['image'])

        input_issuer = json.loads(test_components['2_0_basic_issuer'])
        input_issuer['publicKey'] = input_assertion['verification']['creator']

        key = RSA.generate(2048)
        public_key_pem = key.publickey().export_key()

        cryptographic_key_doc = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'id': input_assertion['verification']['creator'],
            'type': 'CryptographicKey',
            'owner': input_issuer['id'],
            'publicKeyPem': public_key_pem.decode()
        }

        set_up_context_mock()
        for doc in [input_badgeclass, input_issuer, cryptographic_key_doc]:
            responses.add(responses.GET, doc['id'], json=doc, status=200)

        signature = jws.sign(input_assertion, key, algorithm='RS256')

        response = verify(signature, use_cache=False)

        self.assertTrue(response['report']['valid'])

    @responses.activate
    def test_can_full_verify_with_revocation_check(self):
        input_assertion = json.loads(test_components['2_0_basic_assertion'])
        input_assertion['verification'] = {'type': 'signed', 'creator': 'http://example.org/key1'}
        set_up_image_mock(u'https://example.org/beths-robot-badge.png')

        input_badgeclass = json.loads(test_components['2_0_basic_badgeclass'])
        set_up_image_mock(input_badgeclass['image'])

        revocation_list = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'id': 'http://example.org/revocationList',
            'type': 'RevocationList',
            'revokedAssertions': []}
        input_issuer = json.loads(test_components['2_0_basic_issuer'])
        input_issuer['revocationList'] = revocation_list['id']
        input_issuer['publicKey'] = input_assertion['verification']['creator']

        key = RSA.generate(2048)
        public_key_pem = key.publickey().export_key()

        cryptographic_key_doc = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'id': input_assertion['verification']['creator'],
            'type': 'CryptographicKey',
            'owner': input_issuer['id'],
            'publicKeyPem': public_key_pem.decode()
        }

        set_up_context_mock()
        for doc in [input_assertion, input_badgeclass, input_issuer, cryptographic_key_doc, revocation_list]:
            responses.add(responses.GET, doc['id'], json=doc, status=200)

        header = {'alg': 'RS256'}
        payload = json.dumps(input_assertion).encode()
        signature = jws.sign(input_assertion, key, algorithm='RS256')

        response = verify(signature, use_cache=False)

        self.assertTrue(response['report']['valid'])

    @responses.activate
    def test_revoked_badge_marked_invalid(self):
        input_assertion = json.loads(test_components['2_0_basic_assertion'])
        input_assertion['verification'] = {'type': 'signed', 'creator': 'http://example.org/key1'}

        input_badgeclass = json.loads(test_components['2_0_basic_badgeclass'])

        revocation_list = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'id': 'http://example.org/revocationList',
            'type': 'RevocationList',
            'revokedAssertions': [
                {'id': input_assertion['id'], 'revocationReason': 'A good reason, for sure'},
                {'id': 'urn:uuid:52e4c6b3-8c13-4fa8-8482-a5cf34ef37a9'},
                'urn:uuid:6deb4a00-ebce-4b28-8cc2-afa705ef7be4'
            ]
        }
        input_issuer = json.loads(test_components['2_0_basic_issuer'])
        input_issuer['revocationList'] = revocation_list['id']
        input_issuer['publicKey'] = input_assertion['verification']['creator']

        key = RSA.generate(2048)
        public_key_pem = key.publickey().export_key()

        cryptographic_key_doc = {
            '@context': OPENBADGES_CONTEXT_V2_URI,
            'id': input_assertion['verification']['creator'],
            'type': 'CryptographicKey',
            'owner': input_issuer['id'],
            'publicKeyPem': public_key_pem.decode()
        }

        set_up_context_mock()
        for doc in [input_assertion, input_badgeclass, input_issuer, cryptographic_key_doc, revocation_list]:
            responses.add(responses.GET, doc['id'], json=doc, status=200)

        signature = jws.sign(input_assertion, key, algorithm='RS256')

        response = verify(signature, use_cache=False)
        self.assertFalse(response['report']['valid'])
        msg = [a for a in response['report']['messages'] if a.get('name') == VERIFY_SIGNED_ASSERTION_NOT_REVOKED][0]
        self.assertIn('A good reason', msg['result'])

        # Assert pruning went well to eliminate revocationlist revokedAssertions except for the revoked one
        rev_list = [n for n in response['graph'] if n.get('id') == revocation_list['id']][0]
        self.assertEqual(len(rev_list['revokedAssertions']), 1)
        self.assertEqual(rev_list['revokedAssertions'], [revocation_list['revokedAssertions'][0]])


class GraphScrubbingTests(unittest.TestCase):
    def test_can_scrub_revocationlist_from_graph(self):
        assertion_data = {
            'id': 'urn:uuid:339e3b02-8b45-4948-b8d9-3589ec0f75d4',
            'type': 'Assertion',
            'badge': 'urn:uuid:6cd431ae-3155-4a81-b404-3bdc4d64b10d'
        }
        badgeclass_data = {
            'id': assertion_data['badge'],
            'type': 'BadgeClass',
            'issuer': 'http://example.org/issuer'
        }
        issuer_data = {
            'id': 'http://example.org/issuer',
            'type': 'Issuer',
            'revocationList': 'http://example.org/revocations'
        }
        revocation_list = {
            'id': 'http://example.org/revocations',
            'revokedAssertions': [
                'urn:uuid:1fac23b6-f1f7-46bd-9247-b1a94c3f8aaa',
                'urn:uuid:486590d2-89ec-4f71-9c1e-63d1be796a95',
                {'uid': 'abc123'}
            ]
        }
        graph = [assertion_data, badgeclass_data, issuer_data, revocation_list]

        action = patch_node(revocation_list['id'], {'revokedAssertions': []})
        new_graph = graph_reducer(graph, action)

        self.assertEqual(len(new_graph), 4)
        rev_list = [n for n in new_graph if n['id'] == revocation_list['id']][0]
        self.assertEqual(rev_list['revokedAssertions'], [])



from base64 import b64encode
from Crypto.PublicKey import RSA
import json
from jose import jws
import os
from pydux import create_store
import responses
import unittest
import sys

from openbadges.verifier.actions.input import set_input_type, store_input
from openbadges.verifier.actions.action_types import SET_VALIDATION_SUBJECT, STORE_INPUT, STORE_ORIGINAL_RESOURCE
from openbadges.verifier.actions.tasks import add_task
from openbadges.verifier.tasks.task_types import DETECT_INPUT_TYPE
from openbadges.verifier.reducers import main_reducer
from openbadges.verifier.state import INITIAL_STATE
from openbadges.verifier.tasks import run_task
from openbadges.verifier.tasks.input import detect_input_type, process_baked_resource
from openbadges.verifier.tasks.task_types import FETCH_HTTP_NODE, PROCESS_BAKED_RESOURCE
from openbadges.verifier.utils import MESSAGE_LEVEL_ERROR
from openbadges_bakery import bake

try:
    from tests.testfiles.test_components import test_components
    from tests.utils import set_up_context_mock
except (ImportError, SystemError):
    from .testfiles.test_components import test_components
    from .utils import set_up_context_mock


class InputReducerTests(unittest.TestCase):
    def setUp(self):
        self.store = create_store(main_reducer, INITIAL_STATE)

    def test_store_input(self):
        self.store.dispatch(store_input("http://example.com/url1"))
        self.assertEqual(self.store.get_state().get('input').get('value'), 'http://example.com/url1')

    def test_set_input_type(self):
        self.store.dispatch(store_input("http://example.com/url1"))
        self.store.dispatch(set_input_type('url'))
        self.assertEqual(self.store.get_state().get('input').get('input_type'), 'url')


class InputTaskTests(unittest.TestCase):
    def test_input_url_type_detection(self):
        """
        The detect_input_type task should successfully detect
        """
        url = 'http://example.com/assertionmaybe'
        state = INITIAL_STATE.copy()
        state['input']['value'] = url

        success, message, actions = detect_input_type(state, {})

        self.assertTrue(success)
        self.assertEqual(len(actions), 3)
        self.assertEqual(actions[0]['type'], 'SET_INPUT_TYPE')
        self.assertEqual(actions[1]['url'], url)

    def test_input_jsonld_type_detection_replaces_with_url(self):
        """
        The detect_input_type task should successfully detect JSONLD with an id URL
        as input and switch to using an 'id' as URL value if possible
        """
        json_input = test_components['2_0_basic_assertion']
        state = INITIAL_STATE.copy()
        state['input']['value'] = json_input

        success, message, actions = detect_input_type(state, {})

        self.assertTrue(success)
        self.assertEqual(len(actions), 4)
        self.assertEqual(actions[0]['type'], 'STORE_INPUT')
        self.assertEqual(actions[1]['type'], 'SET_INPUT_TYPE')
        self.assertEqual(actions[2]['type'], 'ADD_TASK')
        self.assertEqual(actions[2]['name'], 'FETCH_HTTP_NODE')

        self.assertEqual(actions[0]['input'], actions[2]['url'])
        self.assertEqual(json.loads(json_input)['id'], actions[2]['url'])

    def test_input_jsonld_type_detection_preserves_json(self):
        """
        If the detect_input_type_task can't find an id field as a URL, it preserves
        the input as json
        """
        assertion_dict = json.loads(test_components['2_0_basic_assertion'])
        assertion_dict['id'] = assertion_dict['badge'] = 'urn:org:example:badges:robotics:beth'
        json_input = json.dumps(assertion_dict)
        state = INITIAL_STATE.copy()
        state['input']['value'] = json_input

        success, message, actions = detect_input_type(state, {})

        self.assertTrue(success)
        self.assertEqual(len(actions), 2)
        self.assertEqual(actions[0]['type'], 'SET_INPUT_TYPE')
        self.assertEqual(actions[0]['input_type'], 'json')

    def test_input_json_bad_0_5(self):
        input_data = {
            "recipient": "sha256$ef1253755797c2a3dfd3ab455a7a2080cb9160f2f1fbbf99c475347af1ecc598",
            "issued_on": "2012-12-28",
            "badge": {
                "name": "Completed Rails for Zombies Redux",
                "image": "https://d1ffx7ull4987f.cloudfront.net/images/achievements/large_badge/133/completed-rails-for-zombies-redux-0f73c361c3d5070ca2fa7951e65cbf39.png",
                "description": "Awarded for the completion of Rails for Zombies Redux",
                "version": "0.5.0",
                "criteria": "https://www.codeschool.com/users/mrmickca/badges/133",
                "issuer": {
                    "origin": "http://www.codeschool.com",
                    "org": None,
                    "contact": None,
                    "name": "Code School"
                }
            },
            "salt": "6abf7e9504d73363bdcf9336056f5235",
        }
        input_string = json.dumps(input_data)
        state = INITIAL_STATE
        state['input']['value'] = input_string
        task_meta = add_task(DETECT_INPUT_TYPE)
        result, message, actions = detect_input_type(state, task_meta)
        self.assertTrue(result)
        self.assertEqual(actions[0]['input_type'], 'json')
        self.assertEqual(actions[1]['messageLevel'], MESSAGE_LEVEL_ERROR)

        # 1.0 style hosted JSON input
        input_data['verify'] = {"url": "http://example.org/assertion/1", "type": "hosted"}
        input_data['badge'] = 'http://example.org/badge/1'
        input_string = json.dumps(input_data)
        state['input']['value'] = input_string

        result, message, actions = detect_input_type(state, task_meta)
        self.assertTrue(result)
        self.assertEqual(actions[0]['type'], STORE_INPUT)
        self.assertEqual(actions[0]['input'], input_data['verify']['url'])
        self.assertEqual(actions[1]['input_type'], 'url')
        self.assertEqual(actions[2]['url'], input_data['verify']['url'])  # FETCH_HTTP_NODE
        self.assertEqual(actions[3]['node_id'], input_data['verify']['url'])  # SET_VALIDATION_SUBJECT


class InputJwsTests(unittest.TestCase):
    def setUp(self):
        self.private_key = RSA.generate(2048)
        self.signing_key_doc = {
            'id': 'http://example.org/key1',
            'type': 'CryptographicKey',
            'owner': 'http://example.org/issuer',
            'publicKeyPem': self.private_key.publickey().exportKey()
        }
        self.issuer_data = {
            'id': 'http://example.org/issuer',
            'publicKey': 'http://example.org/key1'
        }
        self.badgeclass = {
            'id': '_:b1',
            'issuer': 'http://example.org/issuer'
        }
        self.verification_object = {
            'id': '_:b0',
            'type': 'SignedBadge',
            'creator': 'http://example.org/key1'
        }
        self.assertion_data = {
            'id': 'urn:uuid:bf8d3c3d-fe60-487c-87a3-06440d0d0163',
            'verification': '_:b0',
            'badge': '_:b1'
        }

        self.signed_assertion = jws.sign(self.assertion_data, self.private_key, algorithm='RS256')

        self.state = {
            'graph': [self.signing_key_doc, self.issuer_data, self.badgeclass,
                      self.verification_object, self.assertion_data]
        }

    @responses.activate
    def test_detect_jws_signed_input_type(self):
        set_up_context_mock()
        # responses.add(responses.GET, badgeclass_data['id'], json=badgeclass_data, status=200)
        # responses.add(responses.GET, issuer_data['id'], json=issuer_data, status=200)
        # responses.add(responses.GET, signing_key['id'], json=signing_key, status=200)

        state = INITIAL_STATE.copy()
        state['input']['value'] = self.signed_assertion

        success, message, actions = detect_input_type(state, {})

        self.assertTrue(success)
        self.assertEqual(len(actions), 2)
        self.assertEqual(actions[0]['input_type'], 'jws')


class InputImageUrlTests(unittest.TestCase):
    def test_input_type_detection(self):
        state = INITIAL_STATE.copy()
        url = 'http://example.org/assertion/1'
        state['input'] = {'value': url}
        task_meta = add_task(DETECT_INPUT_TYPE)
        result, message, actions = detect_input_type(state, task_meta)

        fetch_action = [a for a in actions if a.get('name') == FETCH_HTTP_NODE][0]
        set_action = [a for a in actions if a.get('type') == SET_VALIDATION_SUBJECT][0]

        self.assertTrue(fetch_action.get('is_potential_baked_input'))
        self.assertEqual(set_action.get('node_id'), url)

        task_meta = add_task(DETECT_INPUT_TYPE, is_potential_baked_input=False)
        result, message, actions = detect_input_type(state, task_meta)
        self.assertTrue(result)
        fetch_action = [a for a in actions if a.get('name') == FETCH_HTTP_NODE][0]
        self.assertFalse(fetch_action.get('is_potential_baked_input'))

    @responses.activate
    def test_fetch_task_handles_potential_baked_input(self):
        set_up_context_mock()
        assertion_url = 'http://example.org/assertion/1'
        image_url = 'http://example.org/image'

        with open(os.path.join(os.path.dirname(__file__), 'testfiles', 'public_domain_heart.png'), 'rb') as f:
            baked_file = bake(f, assertion_url)

        responses.add(responses.GET, image_url, body=baked_file.read(), status=200, content_type='image/png')

        task = add_task(FETCH_HTTP_NODE, url=image_url, is_potential_baked_input=True, depth=0)
        result, message, actions = run_task({}, task)

        self.assertTrue(result)
        store_resource_action = [a for a in actions if a.get('type') == STORE_ORIGINAL_RESOURCE][0]
        process_baked_input_action = [a for a in actions if a.get('name') == PROCESS_BAKED_RESOURCE][0]

        self.assertEqual(store_resource_action.get('node_id'), image_url)
        self.assertEqual(process_baked_input_action.get('node_id'), image_url)

        task = add_task(FETCH_HTTP_NODE, url=image_url, is_potential_baked_input=False, depth=0)
        result, message, actions = run_task({}, task)
        self.assertTrue(result)

    def test_process_baked_resource(self):
        image_url = 'http://example.org/image'

        # Store baked Base64 data generated in test_fetch_task_handles_potential_baked_input as a original_resource
        state = {
            'input': {
                'original_json': {
                    image_url: b'data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABUAAAATCAYAAAB/TkaLAAAALWlUWHRvcGVuYmFkZ2VzAAAAAABodHRwOi8vZXhhbXBsZS5vcmcvYXNzZXJ0aW9uLzFq+wCZAAAABmJLR0QA/wD/AP+gvaeTAAAACXBIWXMAAAsTAAALEwEAmpwYAAACtUlEQVQ4y6XRS08TURjG8eec6UxbekmvtFAU0pIajFANxhKNQV2gRF24dOFnUFE3BiMuJUSXLvwCxhhjNCExCiaQECJSqIgXlBgNhd6k9D5tZ+a4kBrAVpH+V5M5b37vyQzBhlivv+mtKCzyILxQpyO6FgdLx9ZWvQ+f2Mozn3qOfSWK1Ojq7sp+mwiYjG3NSVc64if3hxfKM6T8EOw9+SW3HPF4PE5IggBZUcBrBQhWE0rUmDY+frov6m4KClqNCXo1ZL0hSwtFndpsRI4CKl79oGHw3vnf6Mdzp0ZCn5dOeJ1WJNYyMbkk2cvLOIEPKyXJXG/XXxeLzJ9K5/xMVprL55Tjfljb3VbFbITLZrrG9d8ZIgAw2rlf8TY6WOx7WAIgoHLLABqrnMnOIz4qO6xounmXktc+3zMB5AzAZhhwADtMpdWEXN0drkKBzdGcLPVoCGG1gAAg5UVXaTUJScXtpSXGVLVgG1MkBcVYnFKZSCVRlggApVaTEgrIDLQBmhtxKOAICdYiauvNOU4uQVcnLBIAeNTmVTy8hoCQEADXDswVZ4uzIbQUx8HpaUIBoOvDwm7GMKal9DmAlf/RCBC2W4xj1OWY7zx+6ML6u18xgEy3tw/oeX6XKMu9DHBuw4zYLIbRYjK36g4E+ghQBAC6YSPrnJsbCBfFFYHjhgFE/gFGrWbjKzGRSbgDgStlcBNahrvfve+PFsWIhuOHAUSrgDGbxTgiJjOJ1tnZPgIUtjh/xgAy3tFx20GpPcfYaQD2zaDhZXYtm9ozM3NxK1gVXYfphM83aCXEkmfsLAAbgLjVYnhRTIlJz/SbywQQq/y86jFAPe7zDdUTosszdthsMQRKGSnROjXZV+mGFb9phY2Fo8Hg1SVFypv0Wian8snWqclLfwO3HQPU837/LQbw25n/Cf7NEg3as9WWAAAAAElFTkSuQmCC'
                }
            }
        }
        task_meta = add_task(PROCESS_BAKED_RESOURCE, node_id=image_url, depth=0)
        result, message, actions = process_baked_resource(state, task_meta)
        self.assertTrue(result)
        self.assertEqual(len(actions), 2)
        self.assertEqual(actions[0]['type'], STORE_INPUT, "The baked URL is saved")
        self.assertEqual(actions[0]['input'], 'http://example.org/assertion/1', "The baked URL is saved in store_input")
        self.assertEqual(actions[1]['name'], DETECT_INPUT_TYPE)
        self.assertFalse(actions[1]['is_potential_baked_input'])

        state['input']['original_json'][image_url] = 'data:image/jpeg;base64,/9j/4AA'
        result, message, actions = process_baked_resource(state, task_meta)
        self.assertFalse(result, "Task fails when image is not of known type.")

        state['input']['original_json'][image_url] = 'data:image/jpeg;base64,'
        result, message, actions = process_baked_resource(state, task_meta)
        self.assertFalse(result, "Task fails when there is no content.")
        self.assertIn('Cannot determine image type or content', message)



"""
Custom exceptions for the function of Open Badges verification
"""

class SkipTask(Exception):
    """
    This exception indicates that the present task is not ready to be executed
    and should be shuffled to the back of the stack.
    """
    pass


class TaskPrerequisitesError(Exception):
    """
    This exception indicates that the present task has prerequisites better
    expressed in other tasks that were not met in order to run.
    """
    pass


class ValidationError(Exception):
    """
    This exception is used in tasks to indicate that a requirement has not been met.
    """
    def __init__(self, message=None):
        self.message = message



import six

from .actions.action_types import REPORT_MESSAGE
from .utils import list_of, MESSAGE_LEVEL_ERROR, MESSAGE_LEVEL_INFO, MESSAGE_LEVEL_WARNING

INITIAL_STATE = {
    'input': {},
    'graph': [],
    'tasks': [],
    'validationReport': {}
}


# Tasks
def filter_active_tasks(state):
    tasks = state.get('tasks')

    def _task_is_ready(task):
        # Return True if task is not complete and task has no unfulfilled prerequisites
        if task.get('complete'):
            return False

        prerequisites = list_of(task.get('prerequisites', []))
        for prereq in prerequisites:
            prereq_tasks = [pt for pt in tasks if pt.get('name') == prereq or pt.get('task_key') == prereq]
            if not prereq_tasks or not all(task.get('complete') for task in prereq_tasks):
                return False

        return True

    return [t for t in tasks if _task_is_ready(t)]


def filter_failed_tasks(state):
    return [t for t in state.get('tasks') if not t.get('success')]


# Messages
def filter_messages_for_report(state):
    messages = []
    for t in state.get('tasks'):
        if t.get('success') is False or t.get('messageLevel') == MESSAGE_LEVEL_INFO or t.get('name') == REPORT_MESSAGE:
            messages.append(t)
    return messages


def format_message(task_meta):
    ret = {
        'name': task_meta.get('name'),
        'success': task_meta.get('success', False),
        'result': task_meta.get('result', ''),
        'messageLevel': task_meta.get('messageLevel', MESSAGE_LEVEL_ERROR),
    }
    if task_meta.get('node_id'):
        ret['node_id'] = task_meta['node_id']
    if task_meta.get('prop_name'):
        ret['prop_name'] = task_meta['prop_name']
    if not task_meta.get('complete') and not ret['result']:
        ret['result'] = 'Task could not execute.'

    return ret


# Graph
def get_node_by_id(state, node_id):
    """
    Filter state to return first node that matches the requested id.
    :param state: state object with "graph" property as a list
    :param node_id: IRI-format string
    Raises IndexError if no node found.
    """
    return [node for node in state['graph'] if node.get('id') == node_id][0]


def get_node_by_path(state, node_path):
    """
    Filter state to return first node that matches the requested id and property path.
    A path takes the format ['_:b0', 'prop_name', 1, 'another_prop_name'].
    Each entry is either a string (dict key) or non-negative integer (list index).
    :param state: state object with "graph" property as a list
    :param node_path: node path list
    Raises IndexError if no node found or no list index found.
    Raises KeyError if no property found in node.
    Raises TypeError if node next path not a string or list next path not an int.
    """
    if len(node_path) > 1:
        paths = iter(node_path)
        try:
            node_id = next(paths)
            node = get_node_by_id(state, node_id)
            while True:
                prop_name = next(paths)
                if not isinstance(prop_name, six.string_types):
                    raise TypeError(
                        'Node property {} should be a string type to use in a path'.format(prop_name))

                val = node[prop_name]

                if isinstance(val, list):
                    i = next(paths)
                    val = val[i]

                if isinstance(val, dict):
                    node = val
                    continue

                return get_node_by_path(state, [val] + list(paths))  # Recurse with remaining path
        except StopIteration:
            if isinstance(val, (dict, list)):
                return val
            raise TypeError("Node path {} not properly formed.")

    else:
        return get_node_by_id(state, node_path[0])


def node_match_exists(state, node_id):
    try:
        get_node_by_id(state, node_id)
    except IndexError:
        return False
    return True



from .action_types import SET_INPUT_TYPE, SET_PROFILE_ID, STORE_INPUT, STORE_ORIGINAL_RESOURCE


def store_input(badge_input):
    """
    Emits an action that encapsulates user input provided and instructs that it
    be stored in the state.
    :param badge_input: string
    :return: dict
    """
    return {
        'type': STORE_INPUT,
        'input': badge_input
    }


def set_input_type(type_string):
    """
    Emits an action that indicates the type of input provided.
    Options: "url", "json", "jws"
    :param input_type: string
    :return: dict
    """
    if type_string not in ['file', 'json', 'jws', 'url']:
        raise TypeError("Only 'file', 'json', 'jws' or 'url' input types supported.")

    return {
        'type': SET_INPUT_TYPE,
        'input_type': type_string
    }


def store_expected_profile_id(profile_id):
    return {
        'type': SET_PROFILE_ID,
        'node_id': profile_id
    }


def store_original_resource(node_id, data=None, file=None):
    """
    Store a fetched blob of JSON, a JWS string, or an image file
    :param data: string
    :param file: file-like object
    :param node_id: string
    :return: dict
    """
    return {
        'type': STORE_ORIGINAL_RESOURCE,
        'data': data,
        'file': file,
        'node_id': node_id
    }



import copy

from ..actions.action_types import ADD_NODE, PATCH_NODE, PATCH_NODE_REFERENCE, UPDATE_NODE
from ..state import get_node_by_id
from ..utils import list_of


current_node_number = -1
def get_next_blank_node_id():
    global current_node_number
    current_node_number += 1
    return "_:b{}".format(current_node_number)


def _flatten_node(node, node_id=None):
    if node.get('id') is None:
        node['id'] = node_id or get_next_blank_node_id()
    node_list = []

    for prop in node:
        if isinstance(node[prop], dict):
            prop_id = node[prop].get('id', get_next_blank_node_id())
            node_list.extend(_flatten_node(node[prop], prop_id))
            node[prop] = prop_id
        elif isinstance(node[prop], list):
            current_list = node[prop]
            dict_indices = [i for i in range(len(current_list)) if isinstance(current_list[i], dict)]
            for index in dict_indices:
                prop_id = current_list[index].get('id', get_next_blank_node_id())
                node_list.extend(_flatten_node(current_list[index], prop_id))
                current_list[index] = prop_id

    node_list.append(node)
    return node_list


def graph_reducer(state=None, action=None):
    if state is None:
        state = []

    if action.get('type') == ADD_NODE:
        state = list(state)  # copy state instead of mutating original
        new_node = copy.deepcopy(action.get('data'))
        if not new_node.get('id'):
            new_node['id'] = action.get('node_id', get_next_blank_node_id())
        # new_nodes = _flatten_node(new_node, action.get('node_id'))
        new_nodes = [new_node]
        state.extend(new_nodes)
    elif action.get('type') == UPDATE_NODE:
        existing_node = get_node_by_id({'graph': state}, action.get('node_id'))
        updated_node = action.get('data')
        state = [node for node in state if node is not existing_node]
        state.append(updated_node)
    elif action.get('type') == PATCH_NODE:
        try:
            existing_node = get_node_by_id({'graph': state}, action.get('node_id'))
            updated_node = copy.copy(existing_node)
            updated_node.update(action.get('data'))
            state = [node for node in state if node is not existing_node]
            state.append(updated_node)
        except IndexError:
            pass
    elif action.get('type') == PATCH_NODE_REFERENCE:
        node_path = action['node_path']
        new_node_id = action['new_id']
        try:
            existing_node = get_node_by_id({'graph': state}, node_path[0])
            updated_node = copy.copy(existing_node)
            if len(node_path) > 2:
                raise NotImplementedError("TODO: handle complex node paths!")
            prop = node_path[1]
            updated_node[prop] = new_node_id
            state = [node for node in state if node is not existing_node]
            state.append(updated_node)
        except IndexError:
            pass

    return state



"""
INPUT Tasks:
Process user input
"""
DETECT_INPUT_TYPE = 'DETECT_INPUT_TYPE'
PROCESS_BAKED_RESOURCE = 'PROCESS_BAKED_RESOURCE'

"""
GRAPH Tasks:
Fetch, store, and process nodes in the graph related to validation input.
"""
FETCH_HTTP_NODE = 'FETCH_HTTP_NODE'
INTAKE_JSON = 'INTAKE_JSON'
JSONLD_COMPACT_DATA = 'JSONLD_COMPACT_DATA'
PROCESS_JWS_INPUT = 'PROCESS_JWS_INPUT'


"""
OPENBADGES OBJECT Tasks
"""
UPGRADE_0_5_NODE = 'UPGRADE_0_5_NODE'
UPGRADE_1_0_NODE = 'UPGRADE_1_0_NODE'
UPGRADE_1_1_NODE = 'UPGRADE_1_1_NODE'

UPGRADE_TASKS = [UPGRADE_0_5_NODE, UPGRADE_1_0_NODE, UPGRADE_1_1_NODE]


"""
VALIDATION Tasks:
Ensure data is in good shape for relevant Open Badges objects and links between
objects are sound.
"""
DETECT_AND_VALIDATE_NODE_CLASS = 'DETECT_AND_VALIDATE_NODE_CLASS'
VALIDATE_EXPECTED_NODE_CLASS = 'VALIDATE_EXPECTED_NODE_CLASS'
VALIDATE_RDF_TYPE_PROPERTY = 'VALIDATE_RDF_TYPE_PROPERTY'
VALIDATE_PROPERTY = 'VALIDATE_PROPERTY'

# Class Level Validation Tasks
ASSERTION_TIMESTAMP_CHECKS = 'ASSERTION_TIMESTAMP_CHECKS'
ASSERTION_VERIFICATION_CHECK = 'ASSERTION_VERIFICATION_CHECK'
ASSERTION_VERIFICATION_DEPENDENCIES = 'ASSERTION_VERIFICATION_DEPENDENCIES'
CRITERIA_PROPERTY_DEPENDENCIES = 'CRITERIA_PROPERTY_DEPENDENCIES'
FLATTEN_EMBEDDED_RESOURCE = 'FLATTEN_EMBEDDED_RESOURCE'
IDENTITY_OBJECT_PROPERTY_DEPENDENCIES = 'IDENTITY_OBJECT_PROPERTY_DEPENDENCIES'
IMAGE_VALIDATION = 'IMAGE_VALIDATION'
ISSUER_PROPERTY_DEPENDENCIES = 'ISSUER_PROPERTY_DEPENDENCIES'
VALIDATE_EXTENSION_NODE = 'VALIDATE_EXTENSION_NODE'
VALIDATE_EXTENSION_SINGLE = 'VALIDATE_EXTENSION_SINGLE'
VALIDATE_REVOCATIONLIST_ENTRIES = 'VALIDATE_REVOCATIONLIST_ENTRIES'
VERIFY_JWS = 'VERIFY_JWS'
VERIFY_KEY_OWNERSHIP = 'VERIFY_KEY_OWNERSHIP'


"""
VERIFICATION Tasks:
Ensure values are congruent with one another and that relevant verification
rules in the Open Badges Specification are met.
"""
HOSTED_ID_IN_VERIFICATION_SCOPE = 'HOSTED_ID_IN_VERIFICATION_SCOPE'
VERIFY_RECIPIENT_IDENTIFIER = 'VERIFY_RECIPIENT_IDENTIFIER'
VERIFY_SIGNED_ASSERTION_NOT_REVOKED = 'VERIFY_SIGNED_ASSERTION_NOT_REVOKED'


CLASS_VALIDATION_TASKS = (ASSERTION_TIMESTAMP_CHECKS, ASSERTION_VERIFICATION_CHECK,
                          ASSERTION_VERIFICATION_DEPENDENCIES, CRITERIA_PROPERTY_DEPENDENCIES,
                          FLATTEN_EMBEDDED_RESOURCE, IDENTITY_OBJECT_PROPERTY_DEPENDENCIES,
                          IMAGE_VALIDATION, ISSUER_PROPERTY_DEPENDENCIES,
                          VALIDATE_REVOCATIONLIST_ENTRIES, VERIFY_RECIPIENT_IDENTIFIER)


"""
TRIGGERED CONDITIONS
These conditions may be used as prerequisites in other tasks.
"""
SIGNING_KEY_FETCHED = 'SIGNING_KEY_FETCHED'



import os
import responses
import unittest

from pydux import create_store

from openbadges.verifier.actions.input import store_original_resource
from openbadges.verifier.actions.tasks import add_task, report_message
from openbadges.verifier.reducers import main_reducer
from openbadges.verifier.state import INITIAL_STATE
from openbadges.verifier.tasks import task_named
from openbadges.verifier.tasks.task_types import VALIDATE_PROPERTY
from openbadges.verifier.tasks.validation import ValueTypes
from openbadges.verifier.verifier import call_task, generate_report, verify

from openbadges_bakery import bake

try:
    from tests.testfiles.test_components import test_components
except (ImportError, SystemError):
    from .testfiles.test_components import test_components

from tests.utils import set_up_image_mock


class InitializationTests(unittest.TestCase):
    def test_store_initialization(self):
        def no_op(state, action):
            return state
        store = create_store(no_op, INITIAL_STATE)
        self.assertEqual(store.get_state(), INITIAL_STATE)

    @responses.activate
    def test_verify_function(self):
        url = 'https://example.org/beths-robotics-badge.json'
        responses.add(
            responses.GET, url, body=test_components['2_0_basic_assertion'], status=200,
            content_type='application/ld+json'
        )
        set_up_image_mock('https://example.org/beths-robot-badge.png')
        responses.add(
            responses.GET, 'https://w3id.org/openbadges/v2',
            body=test_components['openbadges_context'], status=200,
            content_type='application/ld+json'
        )
        responses.add(
            responses.GET, 'https://example.org/robotics-badge.json',
            body=test_components['2_0_basic_badgeclass'], status=200,
            content_type='application/ld+json'
        )
        set_up_image_mock(u'https://example.org/robotics-badge.png')
        responses.add(
            responses.GET, 'https://example.org/organization.json',
            body=test_components['2_0_basic_issuer'], status=200,
            content_type='application/ld+json'
        )

        results = verify(url)
        self.assertEqual(results.get('input').get('value'), url)
        self.assertEqual(results.get('input').get('input_type'), 'url')

        self.assertEqual(
            len(results['report']['messages']), 0, "There should be no failing tasks.")

    @responses.activate
    def test_verify_of_baked_image(self):
        url = 'https://example.org/beths-robotics-badge.json'
        png_badge = os.path.join(os.path.dirname(__file__), 'testfiles', 'public_domain_heart.png')
        responses.add(
            responses.GET, url, body=test_components['2_0_basic_assertion'], status=200,
            content_type='application/ld+json'
        )
        set_up_image_mock(u'https://example.org/beths-robot-badge.png')
        responses.add(
            responses.GET, 'https://w3id.org/openbadges/v2',
            body=test_components['openbadges_context'], status=200,
            content_type='application/ld+json'
        )
        responses.add(
            responses.GET, 'https://example.org/robotics-badge.json',
            body=test_components['2_0_basic_badgeclass'], status=200,
            content_type='application/ld+json'
        )
        set_up_image_mock(u'https://example.org/robotics-badge.png')
        responses.add(
            responses.GET, 'https://example.org/organization.json',
            body=test_components['2_0_basic_issuer'], status=200,
            content_type='application/ld+json'
        )

        options = {'max_validation_depth': 3}

        with open(png_badge, 'rb') as image:
            baked_image = bake(image, test_components['2_0_basic_assertion'])
            responses.add(responses.GET, 'https://example.org/baked', body=baked_image.read(), content_type='image/png')
            results = verify(baked_image, None, **options)

        # verify gets the JSON out of the baked image, and then detect_input_type
        # will reach out to the assertion URL to fetch the canonical assertion (thus,
        # we expect this to become an URL input type for the verifier).
        self.assertNotEqual(results, None)
        self.assertEqual(results.get('input').get('value'), url)
        self.assertEqual(results.get('input').get('input_type'), 'url')
        self.assertEqual(len(results['report']['messages']), 0, "There should be no failing tasks.")

        # Verify that the same result occurs when passing in the baked image url.
        another_result = verify('https://example.org/baked')
        self.assertTrue(another_result['report']['valid'])
        self.assertEqual(another_result['report']['validationSubject'], results['report']['validationSubject'])

    # def debug_live_badge_verification(self):
    #     """
    #     Developers: Uncomment this test to run a quick verification check in your debugger.
    #     Because this test method name doesn't start with 'test', it will not be automatically run
    #     even when uncommented.
    #     """
    #     results = verify(
    #         u'REPLACE WITH URL, UNCOMMENT, AND RUN')
    #
    #     self.assertTrue(results['report']['valid'])


class MessagesTests(unittest.TestCase):
    def test_message_reporting(self):
        store = create_store(main_reducer, INITIAL_STATE)
        store.dispatch(report_message('TEST MESSAGE'))

        state = store.get_state()
        self.assertEqual(len(state['tasks']), 1)

        result = generate_report(store)
        self.assertEqual(len(result['report']['messages']), 1)
        self.assertEqual(result['report']['messages'][0]['result'], 'TEST MESSAGE')

    def test_valid_when_no_graph(self):
        state = INITIAL_STATE.copy()
        store = create_store(main_reducer, state)
        result = generate_report(store)
        self.assertFalse(result['report']['valid'])


class ResultReportTests(unittest.TestCase):
    def test_original_json_option(self):
        store = create_store(main_reducer, INITIAL_STATE)
        store.dispatch(store_original_resource('http://example.org/1', '{"data": "test data"}'))

        report = generate_report(store)
        self.assertNotIn('original_json', list(report['input'].keys()))

        report = generate_report(store, {'include_original_json': True})
        self.assertIn('original_json', list(report['input'].keys()))

    @responses.activate
    def test_verify_with_original_json(self):
        url = 'https://example.org/beths-robotics-badge.json'
        responses.add(
            responses.GET, url, body=test_components['2_0_basic_assertion'], status=200,
            content_type='application/ld+json'
        )
        responses.add(
            responses.GET, 'https://w3id.org/openbadges/v2',
            body=test_components['openbadges_context'], status=200,
            content_type='application/ld+json'
        )
        responses.add(
            responses.GET, 'https://example.org/robotics-badge.json',
            body=test_components['2_0_basic_badgeclass'], status=200,
            content_type='application/ld+json'
        )
        responses.add(
            responses.GET, 'https://example.org/organization.json',
            body=test_components['2_0_basic_issuer'], status=200,
            content_type='application/ld+json'
        )

        result = verify(url, include_original_json=True)
        self.assertIn('original_json', list(result['input'].keys()))
        self.assertEqual(len(result['input']['original_json']), 3)
        self.assertIn(url, list(result['input']['original_json'].keys()))


class ExceptionHandlingTests(unittest.TestCase):
    def test_can_print_exception(self):
        state = INITIAL_STATE.copy()
        # Create a state that will trigger an exception
        state['graph'] = [AttributeError("Haha this isn't a dict!")]
        task = add_task(
            VALIDATE_PROPERTY, node_id='http://example.org/1', prop_name='turnips',
            prop_type=ValueTypes.TEXT)
        store = create_store(main_reducer, state)
        store.dispatch(task)

        call_task(task_named(VALIDATE_PROPERTY), store.get_state()['tasks'][0], store)

        state = store.get_state()
        self.assertEqual(len(state['tasks']), 1, 'There is one task in state.')
        task = state['tasks'][0]
        self.assertFalse(task['success'])
        self.assertIn('AttributeError:', task['result'], "assert an AttributeError is formatted as the message.")



import base64
import re
import requests
import requests_cache
import six

from ..actions.input import store_original_resource
from ..actions.tasks import add_task
from ..exceptions import TaskPrerequisitesError
from ..state import get_node_by_id, get_node_by_path

from .task_types import IMAGE_VALIDATION
from .utils import (task_result, abbreviate_value,
                    abbreviate_node_id as abv_node,
                    is_data_uri)

def validate_image(state, task_meta, **options):
    try:
        node_id = task_meta.get('node_id')
        node_path = task_meta.get('node_path')
        prop_name = task_meta.get('prop_name', 'image')
        node_class = task_meta.get('node_class')
        required = bool(task_meta.get('required', False))
        if node_id:
            node = get_node_by_id(state, node_id)
            node_path = [node_id]
        else:
            node = get_node_by_path(state, node_path)

        if options.get('cache_backend'):
            session = requests_cache.CachedSession(
                backend=options['cache_backend'], expire_after=options.get('cache_expire_after', 300))
        else:
            session = requests.Session()
    except (IndexError, TypeError, KeyError):
        raise TaskPrerequisitesError()

    actions = []

    image_val = node.get(prop_name)

    if image_val is None:
        return task_result(not required, "Could not load and validate image in node {}".format(abv_node(node_id, node_path)))
    if isinstance(image_val, six.string_types):
        url = image_val
    elif isinstance(image_val, dict):
        url = image_val.get('id')
    elif isinstance(image_val, list):
        return task_result(False, "many images not allowed")
    else:
        raise TypeError("Could not interpret image property value {}".format(
            abbreviate_value(image_val)
        ))
    if is_data_uri(url):
        if task_meta.get('allow_data_uri', False) is False:
            return task_result(False, "Image in node {} may not be a data URI.".format(abv_node(node_id, node_path)))
        try:
            mimetypes = re.match(r'(?P<scheme>^data):(?P<mimetypes>[^,]{0,}?)?(?P<encoding>base64)?,(?P<data>.*$)', url).group(
                'mimetypes')
            if 'image/png' not in mimetypes and 'image/svg+xml' not in mimetypes:
                raise ValueError("Disallowed filetype")
        except (AttributeError, ValueError,):
            return task_result(
                False, "Data URI image does not declare any of the allowed PNG or SVG mime types in {}".format(
                    abv_node(node_id, node_path))
            )
    elif url:
        existing_file = state.get('input', {}).get('original_json', {}).get(url)
        if existing_file:
            return task_result(True, "Image resource already stored for url {}".format(abbreviate_value(url)))
        else:
            try:
                result = session.get(
                    url, headers={'Accept': 'application/ld+json, application/json, image/png, image/svg+xml'}
                )
                result.raise_for_status()
                content_type = result.headers['content-type']
                encoded_body = base64.b64encode(result.content)
                data_uri = "data:{};base64,{}".format(content_type, encoded_body)

            except (requests.ConnectionError, requests.HTTPError, KeyError):
                return task_result(False, "Could not fetch image at {}".format(url))
            else:
                actions.append(store_original_resource(url, data_uri))

    return task_result(True, "Validated image for node {}".format(abv_node(node_id, node_path)), actions)



import aniso8601
import datetime
import json
import six
import pytz

from ..actions.graph import patch_node
from ..actions.tasks import add_task
from ..exceptions import TaskPrerequisitesError
from ..openbadges_context import OPENBADGES_CONTEXT_V1_URI
from ..state import get_node_by_id
from ..tasks.task_types import JSONLD_COMPACT_DATA, UPGRADE_1_1_NODE
from ..utils import list_of

from .utils import task_result, abbreviate_node_id as abv_node
from .validation import OBClasses, ValueTypes, PrimitiveValueValidator


def _upgrade_datetime(t):
    try:
        return datetime.datetime.fromtimestamp(float(t), pytz.utc).isoformat()
    except (TypeError, ValueError):
        try:
            dt = aniso8601.parse_datetime(t)
            if dt.tzinfo is None:
                dt = pytz.utc.localize(dt)
            return dt.isoformat()
        except (ValueError, TypeError):
            dt = datetime.datetime.strptime(t, '%Y-%m-%d')
            if dt.tzinfo is None:
                dt = pytz.utc.localize(dt)
            return dt.isoformat()


def upgrade_1_1_node(state, task_meta, **options):
    try:
        node_id = task_meta.get('node_id')
        node = get_node_by_id(state, node_id)
        data = node.copy()
    except IndexError:
        raise TaskPrerequisitesError()

    actions = []
    expected_class = task_meta.get('expected_class')

    if expected_class == OBClasses.Assertion or data.get('recipient') is not None:
        # Do assertion upgrades
        dt_validator = PrimitiveValueValidator(ValueTypes.DATETIME)
        issued_on = data.get('issuedOn')
        patch = {}
        if not dt_validator(issued_on) and issued_on is not None:
            try:
                patch['issuedOn'] = _upgrade_datetime(issued_on)
            except TypeError:
                return task_result(False, "Could not interpret datetime {}".format(issued_on))

        if data.get('expires') and not dt_validator(data['expires']):
            try:
                patch['expires'] = _upgrade_datetime(data['expires'])
            except TypeError:
                return task_result(False, "Could not interpret datetime {}".format(data['expires']))

        if patch:
            actions.append(patch_node(node_id, patch))

    elif expected_class == OBClasses.BadgeClass or data.get('criteria') is not None:
        # Do badgeclass upgrades
        if data.get('alignment'):
            a_patch = []
            for a_item in list_of(data['alignment']):
                alignment_fixed = False
                alignment_node = a_item.copy()
                for term, new_term in [('url', 'targetUrl'), ('name', 'targetName'), ('description', 'targetDescription')]:
                    if alignment_node.get(term) and not alignment_node.get(new_term):
                        alignment_fixed = True
                        alignment_node[new_term] = alignment_node.pop(term)
                a_patch.append(alignment_node)
            if alignment_fixed:
                patch = {'alignment': a_patch}
                actions.append(patch_node(node_id, patch))
    if actions:
        return task_result(True, "Node {} upgraded from v1.1 to 2.0".format(node_id), actions)
    else:
        return task_result(True, "Node {} needed no content upgrades from v1.1 to 2.0".format(node_id))


def upgrade_1_0_node(state, task_meta, **options):
    try:
        json_data = task_meta['data']
        node_id = task_meta.get('node_id')
        data = json.loads(json_data)
        expected_class = task_meta.get('expected_class')
    except (KeyError, TypeError, ValueError):
        raise TaskPrerequisitesError()

    actions = []

    if expected_class == OBClasses.Assertion or data.get('recipient') is not None:
        expected_class = OBClasses.Assertion
        # Populate 'id' field
        if data.get('id') is None and node_id is None:
            verification_type = data.get('verify', {}).get('type')
            if verification_type == 'hosted':
                node_id = data['verify'].get('url')
            elif verification_type == 'signed' and data.get('uid') is not None:
                node_id = u'uid:{}'.format(data['uid'])

        if node_id:
            data['id'] = node_id
        else:
            return task_result(False, "Could not determine 'id' for Assertion data to upgrade to v1.1+")

        data['type'] = OBClasses.Assertion

    elif expected_class == OBClasses.BadgeClass or data.get('criteria') is not None:
        expected_class = OBClasses.BadgeClass
        if data.get('id') is None and node_id is None:
            # This should not be the case for 1.0 badges, because we should always fetch them from their hosted URLs.
            return task_result(False, "Could not determine 'id' for BadgeClass to upgrade to v1.1+")

        elif node_id is not None:
            data['id'] = node_id

        data['type'] = OBClasses.BadgeClass

    elif expected_class in [OBClasses.Issuer, OBClasses.Profile] or data.get('url') is not None:
        expected_class = OBClasses.Issuer
        if data.get('id') is None and node_id is None:
            # This should not be the case for 1.0 badges, because we should always fetch them from their hosted URLs.
            return task_result(False, "Could not determine 'id' for Issuer to upgrade to v1.1+")

        elif node_id is not None:
            data['id'] = node_id

        data['type'] = OBClasses.Issuer

    data['@context'] = OPENBADGES_CONTEXT_V1_URI

    compact_action = add_task(
        JSONLD_COMPACT_DATA, node_id=node_id, expected_class=expected_class, data=json.dumps(data),
        source_node_path=task_meta.get('source_node_path'))
    actions.append(compact_action)
    actions.append(add_task(
        UPGRADE_1_1_NODE, node_id=node_id, expected_class=expected_class,
        prerequisites=[compact_action['task_key']]
    ))

    return task_result(True, "Upgraded node {} to 1.1".format(node_id), actions)


def upgrade_0_5_node(state, task_meta, **options):
    try:
        json_data = task_meta['data']
        node_id = task_meta.get('node_id')
        data = json.loads(json_data)
        expected_class = task_meta.get('expected_class')
    except (KeyError, TypeError, ValueError):
        raise TaskPrerequisitesError()

    if expected_class is None:
        expected_class = OBClasses.Assertion

    actions = []

    data['type'] = expected_class
    data['@context'] = OPENBADGES_CONTEXT_V1_URI

    if data.get('id') is None and node_id is None:
        return task_result(False, "Could not determine 'id' for data")
    elif data.get('id') is None and node_id is not None:
        data['id'] = node_id

    if data.get('issued_on') is not None and data.get('issuedOn') is None:
        data['issuedOn'] = data.pop('issued_on')

    if data.get('recipient') and isinstance(data['recipient'], six.string_types):
        recipient_data = {
            'identity': data['recipient'],
            'type': 'email',
            'hashed': bool('@' not in data['recipient'])
        }
        if data.get('salt'):
            recipient_data['salt'] = data.pop('salt')
        data['recipient'] = recipient_data

        data['verify'] = {'type': 'HostedBadge'}

    if expected_class == OBClasses.Assertion or data.get('badge'):
        data['badge']['type'] = OBClasses.BadgeClass

    if data.get('badge', {}).get('issuer'):
        issuer = data['badge']['issuer'].copy()
        issuer['type'] = OBClasses.Profile

        for term, new_term in [('contact', 'email'), ('origin', 'url')]:
            if issuer.get(term) and not issuer.get(new_term):
                try:
                    issuer[new_term] = issuer.pop(term)
                except KeyError:
                    pass

        if issuer.get('name') and issuer.get('org'):
            issuer['name'] = u'{}: {}'.format(issuer['name'], issuer.pop('org'))

        data['badge']['issuer'] = issuer

    compact_action = add_task(
        JSONLD_COMPACT_DATA, node_id=node_id, expected_class=expected_class, data=json.dumps(data),
        source_node_path=task_meta.get('source_node_path'))
    actions.append(compact_action)
    actions.append(add_task(
        UPGRADE_1_1_NODE, node_id=node_id, expected_class=expected_class,
        prerequisites=[compact_action['task_key']]
    ))

    return task_result(True, "Upgraded node {} from 0.5 to 1.1".format(abv_node(node_id)), actions)



from Crypto.PublicKey import RSA
from jose import jwk, jws, exceptions as joseexceptions
import json
import six

from ..actions.graph import patch_node
from ..actions.tasks import add_task
from ..actions.validation_report import set_validation_subject
from ..exceptions import TaskPrerequisitesError
from ..state import get_node_by_id, get_node_by_path
from ..utils import list_of, make_string_from_bytes

from .utils import task_result
from .task_types import (ISSUER_PROPERTY_DEPENDENCIES, INTAKE_JSON, SIGNING_KEY_FETCHED, VERIFY_JWS,
                         VERIFY_KEY_OWNERSHIP, VALIDATE_PROPERTY, VALIDATE_REVOCATIONLIST_ENTRIES,
                         VERIFY_SIGNED_ASSERTION_NOT_REVOKED)
from .validation import OBClasses, ValueTypes


def process_jws_input(state, task_meta, **options):
    try:
        data = task_meta['data']
    except KeyError:
        raise TaskPrerequisitesError()

    node_json = jws.get_unverified_claims(data).decode('utf-8')
    node_data = json.loads(node_json)
    node_id = task_meta.get('node_id', node_data.get('id'))
    depth = task_meta.get('depth')

    actions = [
        add_task(INTAKE_JSON, data=node_json, node_id=node_id, depth=depth),
        add_task(VERIFY_JWS, node_id=node_id, data=data, prerequisites=SIGNING_KEY_FETCHED, depth=depth)
    ]
    if node_id:
        actions.append(set_validation_subject(node_id))
    return task_result(True, "Processed JWS-signed data and queued signature verification task", actions)


def verify_jws_signature(state, task_meta, **options):
    try:
        data = task_meta['data']
        node_id = task_meta['node_id']
        key_node = get_node_by_path(state, [node_id, 'verification', 'creator'])
        public_pem = key_node['publicKeyPem']
        depth = task_meta['depth']
    except (KeyError, IndexError,):
        raise TaskPrerequisitesError()

    actions = [
        add_task(VERIFY_KEY_OWNERSHIP, node_id=node_id, depth=depth),
        add_task(
            VALIDATE_PROPERTY, node_path=[node_id, 'badge', 'issuer'], prop_name='revocationList',
            prop_type=ValueTypes.ID, expected_class=OBClasses.RevocationList, fetch=True, required=False,
            prerequisites=[ISSUER_PROPERTY_DEPENDENCIES], depth=depth
        ),
    ]

    key = RSA.import_key(public_pem)
    jwkkey = jwk.construct(key, 'RS256').to_dict()

    try:
        jws.verify(data, jwkkey, None)
    except (joseexceptions.JWSError, joseexceptions.JWSSignatureError,) as e:
        return task_result(
            False, "Signature for node {} failed verification".format(node_id) + " :: " + str(e), actions)

    return task_result(
        True, "Signature for node {} passed verification".format(node_id), actions)


def verify_key_ownership(state, task_meta, **options):
    try:
        node_id = task_meta['node_id']
        issuer_node = get_node_by_path(state, [node_id, 'badge', 'issuer'])
        key_node = get_node_by_path(state, [node_id, 'verification', 'creator'])
        key_id = key_node['id']
        depth = task_meta['depth']
    except (KeyError, IndexError,):
        raise TaskPrerequisitesError()

    actions = []
    if issuer_node.get('revocationList'):
        actions.append(add_task(
            VERIFY_SIGNED_ASSERTION_NOT_REVOKED,
            node_id=node_id,
            prerequisites=[VALIDATE_REVOCATIONLIST_ENTRIES],
            depth=depth
        ))

    issuer_keys = list_of(issuer_node.get('publicKey'))
    if key_id not in issuer_keys:
        return task_result(
            False,
            "Assertion signed by a key {} other than those authorized by issuer profile".format(key_id),
            actions)

    return task_result(
        True, "Assertion signing key {} is properly declared in issuer profile".format(key_id), actions)


def verify_signed_assertion_not_revoked(state, task_meta, **options):
    try:
        assertion_id = task_meta['node_id']
        issuer = get_node_by_path(state, [assertion_id, 'badge', 'issuer'])
    except (IndexError, KeyError, TypeError,):
        raise TaskPrerequisitesError()

    if not issuer.get('revocationList'):
        return task_result(True, 'Assertion {} is not revoked. Issuer {} has no revocation list'.format(
            assertion_id, issuer.get('id')
        ))

    revocation_list = get_node_by_id(state, issuer['revocationList'])
    revoked_assertions = revocation_list['revokedAssertions']

    def _is_match(term, container):
        if isinstance(container, six.string_types):
            return term == container
        return container.get('id') == term

    revoked_match = [a for a in revoked_assertions if _is_match(assertion_id, a)]

    actions = [patch_node(revocation_list['id'], {'revokedAssertions': revoked_match})]

    if len(revoked_match):
        assertion_records = [i for i in state['graph'] if i.get('id') == assertion_id]
        msg = ''
        for a in revoked_match:
            try:
                msg = ' with reason: ' + a['revocationReason']
            except (KeyError, TypeError,):
                continue

        return task_result(False, "Assertion {} has been revoked in RevocationList {}{}".format(
            assertion_id, issuer['revocationList'], msg
        ), actions)

    return task_result(True, "Assertion {} is not marked as revoked in RevocationList {}".format(
        assertion_id, issuer['revocationList']
    ), actions)
